PrimeUI = class(Turbine.UI.Window)
function PrimeUI:Constructor(data, name, x, y, advancedMode, ooc, ic)
    Turbine.UI.Window.Constructor(self)
    self.data = data
    self.width = 0
    self.height = 3000
    self.currentHeight = 0
    self.groupMode = advancedMode
    if self.groupMode then
        self:SetPosition(0, 0)
        self:SetSize(Turbine.UI.Display.GetWidth(), Turbine.UI.Display.GetHeight())
    else
        self:SetPosition(x, y)
    end

    self:SetVisible(true)
    self:SetMouseVisible(false)

    self.ooc = ooc
    self.ic = ic

    self.player = Turbine.Gameplay.LocalPlayer.GetInstance()
    self.playerClass = self.player:GetClass()
    self.ClassAttr = self.player:GetClassAttributes()

    self.skillCallbacks = {}
    self.playerCallbacks = {}
    self.attrCallbacks = {}
    self.effectAddedCallbacks = {}
    self.effectRemovedCallbacks = {}
    self.playerChatCallback = {}
    self.enemyChatCallback = {}
    self.sayChatCallback = {}
    self.partyChatCallback = {}
    self.standardChatCallback = {}
    self.playerHealthCallbacks = {}
    self.healthThresholdCooldowns = {}
    self.healthThresholdPercentage = {}
    self.effectAddedGroupCallbacks = {}
    self.deathChatCallbacks = {}
    self.instanceEventCallbacks = {}
    self.fightEventStartCallbacks = {}
    self.fightEventEndCallbacks = {}
    self.effectTargetAddedCallbacks = {}
    self.healthTargetCallbacks = {}
    self.temporaryHealthTargetCallbacks = {}
    self.powerTargetCallbacks = {}
    self.targetInfoCallbacks = {}
    self.combatCallbacks = {}
    self.playerBenefitCallbacks = {}
    self.playerHealCallbacks = {}

    self.combatState = self.player:IsInCombat()

    self.alignments = {
        ["BottomCenter"] = Turbine.UI.ContentAlignment.BottomCenter,
        ["BottomLeft"] = Turbine.UI.ContentAlignment.BottomLeft,
        ["BottomRight"]	= Turbine.UI.ContentAlignment.BottomRight,
        ["MiddleCenter"] = Turbine.UI.ContentAlignment.MiddleCenter,
        ["MiddleLeft"] = Turbine.UI.ContentAlignment.MiddleLeft,
        ["MiddleRight"] = Turbine.UI.ContentAlignment.MiddleRight,
        ["TopCenter"] = Turbine.UI.ContentAlignment.TopCenter,
        ["TopLeft"] = Turbine.UI.ContentAlignment.TopLeft,
        ["TopRight"] = Turbine.UI.ContentAlignment.TopRight,
    }

    self.fortificationTable = {
        ["Improved Fortification I"] = 1,
        ["Improved Fortification II"] = 2,
        ["Improved Fortification III"] = 3,
        ["Improved Fortification IV"] = 4,
        ["Improved Fortification V"] = 5,
    }

    self.classObjectsToName = {
        [Turbine.Gameplay.Class.Burglar] = "Burglar",
        [Turbine.Gameplay.Class.Captain] = "Captain",
        [Turbine.Gameplay.Class.Champion] = "Champion",
        [Turbine.Gameplay.Class.Guardian] = "Guardian",
        [Turbine.Gameplay.Class.Hunter] = "Hunter",
        [Turbine.Gameplay.Class.LoreMaster] = "LoreMaster",
        [Turbine.Gameplay.Class.Minstrel] = "Minstrel",
        [Turbine.Gameplay.Class.RuneKeeper] = "RuneKeeper",
        [Turbine.Gameplay.Class.Warden] = "Warden",
        [Turbine.Gameplay.Class.Beorning] = "Beorning",
        [Turbine.Gameplay.Class.Brawler] = "Brawler",
        [Turbine.Gameplay.Class.BlackArrow] = "BlackArrow",
        [Turbine.Gameplay.Class.Defiler] = "Defiler",
        [Turbine.Gameplay.Class.Reaver] = "Reaver",
        [Turbine.Gameplay.Class.Stalker] = "Stalker",
        [Turbine.Gameplay.Class.WarLeader] = "WarLeader",
        [Turbine.Gameplay.Class.Weaver] = "Weaver",
    } 

    self.compareFunctions = {
        ["<="] = function(percentage, value) return percentage <= value end,
        [">="] = function(percentage, value) return percentage >= value end,
    }

    self.fortificationID = 0
    self.battleFlowID = 0
    self.venomID = 0

    self.name = name

    -- All tools are contained here
    self.tools = {}
    self.UI = {}
    self.timers = {}
    self.icons = {}
    self.timerCount = 0

    self.ExecuteUpdate = function(delta)    
        --We had an update
        for key,_ in pairs(self.timers) do 
            if key:UpdateText(delta) == true then 
                if key.highlight then key.highlight:Toggle(false) end
                key.currentEffect = ""
                self.timers[key] = nil
                if tablelength(self.timers) == 0 and tablelength(self.icons) == 0 then
                    if self.updateCallback then RemoveCallback(Updater, "ExecuteUpdate", self.updateCallback) end
                    self.updateCallback = nil
                end
            end
        end

        for key,_ in pairs(self.icons) do
            if key:ExecuteUpdate(delta) == true then 
                if key.highlight then key.highlight:Toggle(key.highlightWhileOff) end
                self.icons[key] = nil
                if tablelength(self.timers) == 0 and tablelength(self.icons) == 0 then
                    if self.updateCallback then RemoveCallback(Updater, "ExecuteUpdate", self.updateCallback) end
                    self.updateCallback = nil
                end
            end
        end
    end

    self.Update = function()
        if Turbine.Engine.GetGameTime() > self.reloadTime then
            self.player = Turbine.Gameplay.LocalPlayer.GetInstance()
            self.playerClass = self.player:GetClass()
            self.ClassAttr = self.player:GetClassAttributes()
            self:UnloadObject()
            self:Setup()
            self:SetWantsUpdates(false)
        end
    end

    self.combatVisibility = function()

        if self.player:IsInCombat() then
            if self.ic then
                self:SetVisible(true)
            else
                self:SetVisible(false)
            end
        else
            if self.ooc then
                self:SetVisible(true)
            else
                self:SetVisible(false)
            end

        end
    end
    if not(self.ooc) or not(self.ic) then
        table.insert(self.combatCallbacks, AddCallback(self.player, "InCombatChanged", function()
            self.combatVisibility()
        end))
    end

    self.combatVisibility()
    self:Setup()
end

function PrimeUI:ReloadWithDelay()
    if self.skills  then
        self.reloadTime = Turbine.Engine.GetGameTime() + 5
        self:SetWantsUpdates(true)
    end
end

function PrimeUI:ToggleMove()
    self.moveBox:ToggleMove()
    return self:GetPosition()
end

function PrimeUI:ToggleAdvancedMove()
    for key, value in pairs(self.UI) do
        if value.moveBox then
            value.moveBox:ToggleMove()
            value.savePosition()
        end
    end

    for key, value in pairs(self.tools) do
        if value.moveBox then
            value.moveBox:ToggleMove()
            value.savePosition()
        end
    end
    return self.data
end

function PrimeUI:Setup(newData)

    if newData then self.data = newData end

    self.currentHeight = 0
    self.currentOrder = 0
    self.sameOrderOffset = 0
    self.width = 0

    for toolName, data in spairs(self.data, function(t, a, b) return t[a]["Order"] < t[b]["Order"] end) do
        self:CalculateNewBiggestWidth(data)
    end

    if not self.groupMode then
        self:SetSize(self.width, self.height)
    end


    -- Setup all the tools
    for toolName, data in spairs(self.data, function(t, a, b) return t[a]["Order"] < t[b]["Order"] end) do
        -- Make sure we udpate offset before creating tool
        self.currentHeight = self.currentHeight + (data["OffsetAbove"] or 0)

        -- Setup dynamic skill bar
        if data["Type"] == "Skills" then
            self:SetupSkillBar(data, self)
        end

        if data["Type"] == "MoraleBar" then
            self:SetupMoraleBar(data, self)
        end

        if data["Type"] == "MoraleBar_Target" then
            self:SetupTargetMoraleBar(data, self)
        end

        if data["Type"] == "PowerBar" then
            self:SetupPowerBar(data, self)
        end

        if data["Type"] == "PowerBar_Target" then
            self:SetupTargetPowerBar(data, self)
        end

        if data["Type"] == "WrathBar" then
            self:SetupWrathBar(data, self)
        end

        if data["Type"] == "ResourceBar" then
            self:ResourceBar(data, self)
        end

        if data["Type"] == "ValueBar" then
            self:ValueBar(data, self)
        end

        if data["Type"] == "BuffBar" then
            self:BuffBar(data, self)
        end

        if data["Type"] == "BuffBar_Target" then
            self:TargetBuffBar(data, self)
        end
        
        if data["Type"] == "Icon" then
            self:Icon(data, self)
        end

        if data["Type"] == "TimerIcon" then
            self:TimerIcon(data, self)
        end

        if data["Type"] == "Timer" then
            self:Timer(data, self)
        end

        if data["Type"] == "Text" then
            self:Text(data, self)
        end

        if data["Type"] == "Countdown" then
            self:Countdown(data, self)
        end

        if data["Type"] == "Sound" then
            self:Sound(data, self)
        end

        if data["Type"] == "TimerGroup" then
            self:TimerGroup(data, self)
        end

        if data["Type"] == "SkillLabel" then
            self:SkillLabel(data, self)
        end

        if data["Type"] == "Alert" then
            self:Alert(data, self)
        end

        if data["Type"] == "AlertGroup" then
            self:AlertGroup(data, self)
        end

        if data["Type"] == "Ballad Bar" then
            self:BalladBar(data, self)
        end

        if data["Type"] == "Circle Timer" then
            self:CircleTimer(data, self)
            self:SetRotation({x=0, y=0, z=0})
        end

        if data["Type"] == "Note Event" then
            self:NoteEvent(data, self)
        end

        -- Make sure we udpate offset after creating tool
        self.currentHeight = self.currentHeight + (data["OffsetBelow"] or 0)
    end
    if self.currentHeight < 10 then self.currentHeight = 100 end
    if not self.groupMode then
        self:SetSize(self.width, self.currentHeight)
    end
    self.moveBox = MoveBox(self, (self.width or 100), (self.currentHeight or 100), self.name)
end

function PrimeUI:CalculateNewBiggestWidth(data)
    local width = 100

    if data["Width"] and data["Width"] > 32 then width = data["Width"] end
    if data["BorderSize"] then width = width + 2 * data["BorderSize"] end
    if data["Scale"] then width = width * data["Scale"] end

    if self.width < width then self.width = width end

end

function PrimeUI:ValueBar(data, parent)
    local valuebar = Turbine.UI.Control()
    valuebar:SetParent(parent)
    if self.groupMode then
        valuebar:SetPosition(data["X"], data["Y"])
        valuebar.moveBox = MoveBox(valuebar, data["Width"] + 2 * data["BorderSize"], data["Height"] + 2 * data["BorderSize"], "valuebar")
        valuebar.savePosition = function()
            data["X"], data["Y"] = valuebar:GetPosition()
        end
    else
        valuebar:SetPosition(self:ContentAlignement(data["BarAlignment"], data["Width"] + 2 * data["BorderSize"]), self.currentHeight)
    end
    valuebar:SetSize(data["Width"] + 2 * data["BorderSize"], data["Height"] + 2 * data["BorderSize"])
    valuebar:SetMouseVisible(false)
    valuebar:SetBackColor(self:ColourTupleToTurbineColour(data["BorderColour"]))

    -- args (parent, width, height, x, y, barAlignment, barColour, backgroundColour, maxValue, barName, vertical, afterAnimation)
    valuebar.bar = ValueBar(valuebar, data["Width"], data["Height"], data["BorderSize"], data["BorderSize"], 
    self:AlignmentToNumberSmall(data["Alignment"]), self:ColourTupleToTurbineColour(data["ForegroundColour"]), self:ColourTupleToTurbineColour(data["BackgroundColour"]), 20, data["Texture"], data["Vertical"], nil, data["Visibility_Threshold"])

    if tablelength(data["EffectAdded"]) > 0 then
        table.insert(self.effectAddedCallbacks, AddCallback(PlayerEffectsHandler, "EffectAdded", 
        function(effect)
            local time = data["EffectAdded"][effect:GetName()]
            if time then
                local maxTime = time
                if time == true then
                    time = effect:GetDuration() - (Turbine.Engine.GetGameTime() - effect:GetStartTime())
                    maxTime = effect:GetDuration()
                end
                valuebar.bar:SetMaxValue(maxTime)
                valuebar.bar:SetCurrentValue(time)
            end
        end))
    end

    if tablelength(data["EffectAdded_Target"]) > 0 then
        table.insert(self.effectTargetAddedCallbacks, TargetHandlerObject:AddEffectCallback( 
        function(effect)
            local time = data["EffectAdded_Target"][effect:GetName()]
            if time then
                local maxTime = time
                if time == true then
                    time = effect:GetDuration() - (Turbine.Engine.GetGameTime() - effect:GetStartTime())
                    maxTime = effect:GetDuration()
                end
                valuebar.bar:SetMaxValue(maxTime)
                valuebar.bar:SetCurrentValue(time)
            end
        end))
    end

    if tablelength(data["EffectAdded_Group"]) > 0 then
        table.insert(self.effectAddedGroupCallbacks, AddCallback(GroupEffectsHandler, "EffectAdded", 
        function(args)
            local time = data["EffectAdded_Group"][args.effect:GetName()]
            if time then
                local maxTime = time
                if time == true then
                    time = args.effect:GetDuration() - (Turbine.Engine.GetGameTime() - args.effect:GetStartTime())
                    maxTime = args.effect:GetDuration()
                end
                valuebar.bar:SetMaxValue(maxTime)
                valuebar.bar:SetCurrentValue(time)
            end
        end))
    end

    if tablelength(data["EffectRemoved"]) > 0 then
        table.insert(self.effectRemovedCallbacks, AddCallback(PlayerEffectsHandler, "EffectRemoved", 
        function(effect)
            local time = data["EffectRemoved"][effect:GetName()]
            if time then
                local maxTime = time
                if time == true then
                    time = effect:GetDuration() - (Turbine.Engine.GetGameTime() - effect:GetStartTime())
                    maxTime = effect:GetDuration()
                end
                valuebar.bar:SetMaxValue(maxTime)
                valuebar.bar:SetCurrentValue(time)
            end
        end))
    end

    if tablelength(data["Skills"]) > 0 then
        self.skills = true
        for key, value in pairs(data["Skills"]) do
            if SkillEventHandler:SkillExists(key) then
                -- Add the callback for this skill slot
                if not(self.skillCallbacks[key]) then self.skillCallbacks[key] = {} end
                table.insert(self.skillCallbacks[key], AddCallback(SkillEventHandler, key, function(skill) 
                    local time = value
                    local maxTime = value
                    if time == true then
                        time = skill:GetResetTime() - Turbine.Engine.GetGameTime()
                        maxTime = skill:GetCooldown()
                    end
                    valuebar.bar:SetMaxValue(maxTime)
                    valuebar.bar:SetCurrentValue(time)
                end))
            end
        end
    end

    if tablelength(data["EnemyHits"]) > 0 then
        table.insert(self.enemyChatCallback, AddCallback(ChatEventHandler, "EnemyCombat", function(args)
            local value = data["EnemyHits"][args.SkillName]
            if value then
                valuebar.bar:SetMaxValue(value)
                valuebar.bar:SetCurrentValue(value)
            end
        end))

    end

    if tablelength(data["PlayerHits"]) > 0 then
        table.insert(self.playerChatCallback, AddCallback(ChatEventHandler, "PlayerCombat", function(args)
            local value = data["PlayerHits"][args.SkillName]
            if value then
                valuebar.bar:SetMaxValue(value)
                valuebar.bar:SetCurrentValue(value)
            end
        end))
    end

    if tablelength(data["PlayerHeal"]) > 0 then
        table.insert(self.playerHealCallbacks, AddCallback(ChatEventHandler, "PlayerHeal", function(args)
            local value = data["PlayerHeal"][args.SkillName]
            if value then
                valuebar.bar:SetMaxValue(value)
                valuebar.bar:SetCurrentValue(value)
            end
        end))
    end

    if tablelength(data["PlayerBenefit"]) > 0 then
        table.insert(self.playerBenefitCallbacks, AddCallback(ChatEventHandler, "PlayerBenefit", function(args)
            local value = data["PlayerBenefit"][args.SkillName]
            if value then
                valuebar.bar:SetMaxValue(value)
                valuebar.bar:SetCurrentValue(value)
            end
        end))
    end

    if tablelength(data["StandardChat"]) > 0 then
        table.insert(self.standardChatCallback, AddCallback(ChatEventHandler, "Standard", function(message)
            local value = data["StandardChat"][message]
            if value then
                valuebar.bar:SetMaxValue(value)
                valuebar.bar:SetCurrentValue(value)
            end
        end))
    end

    if tablelength(data["PartyChat"]) > 0 then
        table.insert(self.partyChatCallback, AddCallback(ChatEventHandler, "Party", function(message)
            local value = data["PartyChat"][message]
            if value then
                valuebar.bar:SetMaxValue(value)
                valuebar.bar:SetCurrentValue(value)
            end
        end))
    end

    if tablelength(data["SayChat"]) > 0 then
        table.insert(self.sayChatCallback, AddCallback(ChatEventHandler, "SayLine", function(message)
            if data["SayChat"][message]  then
                valuebar.bar:SetMaxValue(data["SayChat"][message])
                valuebar.bar:SetCurrentValue(data["SayChat"][message])
            end
        end))
    end
    
    self.currentHeight = self.currentHeight + data["Height"] + 2 * data["BorderSize"]

    table.insert(self.UI, valuebar)
    table.insert(self.tools, valuebar.bar)
end

function PrimeUI:TargetBuffBar(data, parent)
    local x = self:ContentAlignement(data["Alignment"], data["Width"])
    local y = self.currentHeight
    if self.groupMode then
        x = data.X
        y = data.Y
    end

    local buffbar = DynamicSkillBar(parent, x, y, data["Width"], 32 * data["Scale"], 32 * data["Scale"], self:AlignmentToNumber(data["Alignment"]), true, false, data["Offset"])
    buffbar.name = ""
    buffbar.uniqueTable = {}
    buffbar.uniqueTableIDs = {}

    if self.groupMode then
        buffbar.moveBox = MoveBox(buffbar, data["Width"], data["Width"], "Target Buffbar")
        buffbar.savePosition = function()
            data["X"], data["Y"] = buffbar:GetPosition()
        end
    end

    if tablelength(data["EffectAdded_Target"]) > 0 or data["AllEffects"] then
        table.insert(self.effectTargetAddedCallbacks, TargetHandlerObject:AddEffectCallback(
        function(effect)
            local name = effect:GetName()
            if data["EffectAdded_Target"][name] or data["AllEffects"] then
                if (not(data["OnlyBuffs"]) and not(data["OnlyDebuffs"])) or (data["OnlyDebuffs"] and effect:IsDebuff()) or (data["OnlyBuffs"] and not(effect:IsDebuff())) then
                    local highlight = data["AllHighlight"]
                    local icon = nil
                    local value = data["EffectAdded_Target"][name]
                    if value then
                        highlight = highlight or value["Highlight"]
                        if value["Icon"] == true then 
                            icon = effect:GetIcon()
                        else
                            icon = value["Icon"] 
                        end

                        local time = value["Time"]
                        if time == true then
                            time = effect:GetDuration() - (Turbine.Engine.GetGameTime() - effect:GetStartTime())
                        end

                        if value["Fit_Icon"] and value["Icon"] and value["Icon"] ~= true then
                            if effect:GetIcon() ~= tonumber(value["Icon"]) then 
                                return 
                            end
                        end

                        if value["Unique"] then
                            local id = string.match(name, "%D+") .. tostring(icon)
                            if not(buffbar.uniqueTable[id]) then
                                buffbar.uniqueTable[id] = Turbine.Engine.GetGameTime() + time - (Turbine.Engine.GetGameTime() - effect:GetStartTime())
                                buffbar.uniqueTableIDs[id] = effect:GetID()
                            elseif Turbine.Engine.GetGameTime() < buffbar.uniqueTable[id] then
                                buffbar:RemoveLabel(buffbar.uniqueTableIDs[id])
                                local ID = effect:GetID()
                                buffbar:AddLabel(ID, ClassTimerLabel(buffbar, 0, 0, 32, 32, icon, false, false, nil, 1, 1, false, false, true, data["Scale"], name, true, self.alignments[data["TextAlignment"]], self.alignments[data["StackTextAlignment"]], highlight, self:FontToFont(data["Font_Timer"]), self:FontToFont(data["Font_Stack"]), self:ColourTupleToTurbineColour(data["Font_Timer_Colour"]), self:ColourTupleToTurbineColour(data["Font_Stack_Colour"])), time - (Turbine.Engine.GetGameTime() - effect:GetStartTime()))
                                buffbar:ToggleLabel(ID, true, time - (Turbine.Engine.GetGameTime() - effect:GetStartTime()))
                                buffbar.uniqueTable[id] = Turbine.Engine.GetGameTime() + time - (Turbine.Engine.GetGameTime() - effect:GetStartTime())
                                buffbar.uniqueTableIDs[id] = effect:GetID()
                                return
                            end
                        end

                        local ID = effect:GetID()
                        buffbar:AddLabel(ID, ClassTimerLabel(buffbar, 0, 0, 32, 32, icon, false, false, nil, 1, 1, false, false, true, data["Scale"], name, true, self.alignments[data["TextAlignment"]], self.alignments[data["StackTextAlignment"]], highlight, self:FontToFont(data["Font_Timer"]), self:FontToFont(data["Font_Stack"]), self:ColourTupleToTurbineColour(data["Font_Timer_Colour"]), self:ColourTupleToTurbineColour(data["Font_Stack_Colour"])), time - (Turbine.Engine.GetGameTime() - effect:GetStartTime()))
                        buffbar:ToggleLabel(ID, true, time - (Turbine.Engine.GetGameTime() - effect:GetStartTime()))
                        return
                    else
                        return
                    end

                    local ID = effect:GetID()
                    buffbar:AddLabel(ID, ClassTimerLabel(buffbar, 0, 0, 32, 32, icon, false, false, nil, 1, 1, false, false, true, data["Scale"], name, true, self.alignments[data["TextAlignment"]], self.alignments[data["StackTextAlignment"]], highlight, self:FontToFont(data["Font_Timer"]), self:FontToFont(data["Font_Stack"]), self:ColourTupleToTurbineColour(data["Font_Timer_Colour"]), self:ColourTupleToTurbineColour(data["Font_Stack_Colour"])), effect:GetDuration())
                    buffbar:ToggleLabel(ID, true, effect:GetDuration() - (Turbine.Engine.GetGameTime() - effect:GetStartTime()))
                end
            end
        end))

        if data["Reset_On_Retarget"] then
            table.insert(self.targetInfoCallbacks, TargetHandlerObject:AddTargetInfoCallback(function(value) 
                if value then
                    if buffbar.name ~= value["Name"] then
                        buffbar:Clear()
                        buffbar.uniqueTable = {}
                        buffbar.uniqueTableIDs = {}
                    end
                    buffbar.name = value["Name"]
                else
                    buffbar.name = ""
                    buffbar:Clear()
                end
                
            end))
        end
    end

    self.currentHeight = self.currentHeight + 32 * data["Scale"]
    table.insert(self.tools, buffbar)
end

function PrimeUI:BuffBar(data, parent)
    local x = self:ContentAlignement(data["Alignment"], data["Width"])
    local y = self.currentHeight
    if self.groupMode then
        x = data.X
        y = data.Y
    end

    local buffbar = DynamicSkillBar(parent, x, y, data["Width"], (data["BorderSize"] and data["BorderSize"] * 2 or 0) + 32 * data["Scale"], (data["BorderSize"] and data["BorderSize"] * 2 or 0) + 32 * data["Scale"], self:AlignmentToNumber(data["Alignment"]), true, true, data["Offset"], true)
    
    buffbar.SkillID = {}
    buffbar.UniqueNames = {}

    if self.groupMode then
        buffbar.moveBox = MoveBox(buffbar, data["Width"], data["Width"], "Buffbar")
        buffbar.savePosition = function()
            data["X"], data["Y"] = buffbar:GetPosition()
        end
    end

    if tablelength(data["Effects"]) > 0 or data["AllEffects"] then
        buffbar.effectFunction = function(effect)
            local name = effect:GetName()
            if data["Effects"][name] or data["AllEffects"] then
                if (not(data["OnlyBuffs"]) and not(data["OnlyDebuffs"])) or (data["OnlyDebuffs"] and effect:IsDebuff()) or (data["OnlyBuffs"] and not(effect:IsDebuff())) then
                    local highlight = data["AllHighlight"]
                    local icon = 0
                    local value = data["Effects"][name]
                    local time = value["Time"]
                    if value then
                        highlight = highlight or value["Highlight"]
                        if value["Icon"] == true then 
                            icon = effect:GetIcon()
                        else
                            icon = value["Icon"] 
                        end

                        if value["Fit_Icon"] and value["Icon"] and value["Icon"] ~= true then
                            if effect:GetIcon() ~= tonumber(value["Icon"]) then 
                                return 
                            end
                        end

                        if time == true then
                            time = effect:GetDuration() - (Turbine.Engine.GetGameTime() - effect:GetStartTime())
                        end
                    else
                        icon = effect:GetIcon()
                    end

                    local ID = effect:GetID()
                    buffbar.SkillID[ID] = true
                    buffbar:AddLabel(ID, ClassTimerLabel(buffbar, 0, 0, 32, 32, icon, false, false, nil, 1, 1, false, false, true, data["Scale"], name, true, self.alignments[data["TextAlignment"]], self.alignments[data["StackTextAlignment"]], highlight, self:FontToFont(data["Font_Timer"]), self:FontToFont(data["Font_Stack"]), self:ColourTupleToTurbineColour(data["Font_Timer_Colour"]), self:ColourTupleToTurbineColour(data["Font_Stack_Colour"]), data["BorderSize"], self:ColourTupleToTurbineColour(data["BorderColour"]), self:ColourTupleToTurbineColour(data["HighlightColour"]), data["HighlightSize"]), time)
                    buffbar:ToggleLabel(ID, true, time)
                end
            end
        end

        for i = 1, self.player:GetEffects():GetCount() do
            buffbar.effectFunction(self.player:GetEffects():Get(i))
        end

        table.insert(self.effectAddedCallbacks, AddCallback(PlayerEffectsHandler, "EffectAdded", buffbar.effectFunction))

        table.insert(self.effectRemovedCallbacks, AddCallback(PlayerEffectsHandler, "EffectRemoved", 
        function(effect)
            local ID = effect:GetID()
            if buffbar.SkillID[ID] then
                buffbar:RemoveLabel(ID)
            end

        end))
    end

    if tablelength(data["Skills"]) > 0 then
        self.skills = true
        for key, value in pairs(data["Skills"]) do
            if SkillEventHandler:SkillExists(key) then
                -- Add the callback for this skill slot
                if not(self.skillCallbacks[key]) then self.skillCallbacks[key] = {} end
                table.insert(self.skillCallbacks[key], AddCallback(SkillEventHandler, key, function(skill) 
                    local icon = value["Icon"]
                    if value["Icon"] == true then
                        icon = skill:GetSkillInfo():GetIconImageID()
                    end

                    local time = value["Time"]
                    if value["Time"] == true then
                        time = skill:GetResetTime() - Turbine.Engine.GetGameTime()
                    end

                    local highlight = data["AllHighlight"] or value["Highlight"]

                    if not(buffbar:LabelExists(key)) then
                        buffbar:AddLabel(key, ClassTimerLabel(buffbar, 0, 0, 32, 32, icon, false, false, nil, 1, 1, false, false, true, data["Scale"], key, true, self.alignments[data["TextAlignment"]], self.alignments[data["StackTextAlignment"]], highlight, self:FontToFont(data["Font_Timer"]), self:FontToFont(data["Font_Stack"]), self:ColourTupleToTurbineColour(data["Font_Timer_Colour"]), self:ColourTupleToTurbineColour(data["Font_Stack_Colour"]), data["BorderSize"], self:ColourTupleToTurbineColour(data["BorderColour"]), self:ColourTupleToTurbineColour(data["HighlightColour"]), data["HighlightSize"]), time)
                    end
                    buffbar:ToggleLabel(key, true, time)
                end))
            end
        end
    end

    if tablelength(data["PlayerHits"]) > 0 then
        for key, value in pairs(data["PlayerHits"]) do
            table.insert(self.playerChatCallback, AddCallback(ChatEventHandler, "PlayerCombat", function(args)
                local value = data["PlayerHits"][args.SkillName]
                if value then  
                    local highlight = data["AllHighlight"] or value["Highlight"]
                    if not(buffbar:LabelExists(key)) then
                        buffbar:AddLabel(key, ClassTimerLabel(buffbar, 0, 0, 32, 32, value["Icon"], false, false, nil, 1, 1, false, false, true, data["Scale"], key, true, self.alignments[data["TextAlignment"]], self.alignments[data["StackTextAlignment"]], highlight, self:FontToFont(data["Font_Timer"]), self:FontToFont(data["Font_Stack"]), self:ColourTupleToTurbineColour(data["Font_Timer_Colour"]), self:ColourTupleToTurbineColour(data["Font_Stack_Colour"]), data["BorderSize"], self:ColourTupleToTurbineColour(data["BorderColour"]), self:ColourTupleToTurbineColour(data["HighlightColour"]), data["HighlightSize"]), value["Time"])
                    end
                    buffbar:ToggleLabel(key, true, value["Time"])
                end
            end))
        end
    end

    if tablelength(data["PlayerHeal"]) > 0 then
        for key, value in pairs(data["PlayerHeal"]) do
            table.insert(self.playerHealCallbacks, AddCallback(ChatEventHandler, "PlayerHeal", function(args)
                local value = data["PlayerHeal"][args.SkillName]
                if value then  
                    local highlight = data["AllHighlight"] or value["Highlight"]
                    if not(buffbar:LabelExists(key)) then
                        buffbar:AddLabel(key, ClassTimerLabel(buffbar, 0, 0, 32, 32, value["Icon"], false, false, nil, 1, 1, false, false, true, data["Scale"], key, true, self.alignments[data["TextAlignment"]], self.alignments[data["StackTextAlignment"]], highlight, self:FontToFont(data["Font_Timer"]), self:FontToFont(data["Font_Stack"]), self:ColourTupleToTurbineColour(data["Font_Timer_Colour"]), self:ColourTupleToTurbineColour(data["Font_Stack_Colour"]), data["BorderSize"], self:ColourTupleToTurbineColour(data["BorderColour"]), self:ColourTupleToTurbineColour(data["HighlightColour"]), data["HighlightSize"]), value["Time"])
                    end
                    buffbar:ToggleLabel(key, true, value["Time"])
                end
            end))
        end
    end

    if tablelength(data["PlayerBenefit"]) > 0 then
        for key, value in pairs(data["PlayerBenefit"]) do
            table.insert(self.playerBenefitCallbacks, AddCallback(ChatEventHandler, "PlayerBenefit", function(args)
                local value = data["PlayerBenefit"][args.SkillName]
                if value then  
                    local highlight = data["AllHighlight"] or value["Highlight"]
                    if not(buffbar:LabelExists(key)) then
                        buffbar:AddLabel(key, ClassTimerLabel(buffbar, 0, 0, 32, 32, value["Icon"], false, false, nil, 1, 1, false, false, true, data["Scale"], key, true, self.alignments[data["TextAlignment"]], self.alignments[data["StackTextAlignment"]], highlight, self:FontToFont(data["Font_Timer"]), self:FontToFont(data["Font_Stack"]), self:ColourTupleToTurbineColour(data["Font_Timer_Colour"]), self:ColourTupleToTurbineColour(data["Font_Stack_Colour"]), data["BorderSize"], self:ColourTupleToTurbineColour(data["BorderColour"]), self:ColourTupleToTurbineColour(data["HighlightColour"]), data["HighlightSize"]), value["Time"])
                    end
                    buffbar:ToggleLabel(key, true, value["Time"])
                end
            end))
        end
    end

    if tablelength(data["EnemyHits"]) > 0 then
        for key, value in pairs(data["EnemyHits"]) do
            table.insert(self.enemyChatCallback, AddCallback(ChatEventHandler, "EnemyCombat", function(args)
                local value = data["EnemyHits"][args.SkillName]
                if value then  
                    local highlight = data["AllHighlight"] or value["Highlight"]
                    if not(buffbar:LabelExists(key)) then
                        buffbar:AddLabel(key, ClassTimerLabel(buffbar, 0, 0, 32, 32, value["Icon"], false, false, nil, 1, 1, false, false, true, data["Scale"], key, true, self.alignments[data["TextAlignment"]], self.alignments[data["StackTextAlignment"]], highlight, self:FontToFont(data["Font_Timer"]), self:FontToFont(data["Font_Stack"]), self:ColourTupleToTurbineColour(data["Font_Timer_Colour"]), self:ColourTupleToTurbineColour(data["Font_Stack_Colour"]), data["BorderSize"], self:ColourTupleToTurbineColour(data["BorderColour"]), self:ColourTupleToTurbineColour(data["HighlightColour"]), data["HighlightSize"]), value["Time"])
                    end
                    buffbar:ToggleLabel(key, true, value["Time"])
                end
            end))
        end
    end

    self.currentHeight = self.currentHeight + (data["BorderSize"] and data["BorderSize"] * 2 or 0) + 32 * data["Scale"]
    table.insert(self.tools, buffbar)

end

function PrimeUI:ResourceBar(data, parent)
    local boxes = 0
    if self.playerClass == Turbine.Gameplay.Class.Champion then
        boxes = 5
    elseif self.playerClass == Turbine.Gameplay.Class.Hunter then
        boxes = 9
    elseif self.playerClass == Turbine.Gameplay.Class.RuneKeeper then
        if data["RKLayout"] then
            boxes = 9
        else
            boxes = 20
        end
    elseif self.playerClass == Turbine.Gameplay.Class.Brawler then
        boxes = 9
    elseif self.playerClass == Turbine.Gameplay.Class.Guardian then
        boxes = 5
    elseif self.playerClass == Turbine.Gameplay.Class.Beorning then
        boxes = 5
    elseif self.playerClass == Turbine.Gameplay.Class.BlackArrow then
        boxes = 5
    elseif self.playerClass == Turbine.Gameplay.Class.Weaver then
        boxes = 5
    end

    local x = self:ContentAlignement(data["Alignment"], data["Width"] + data["BorderSize"] * 2)
    local y = self.currentHeight
    if self.groupMode then
        x = data.X
        y = data.Y
    end

    local resourceBar = ResourceBar(parent, data["Width"], data["Height"], x, y, boxes, 
    self:ColourTupleToTurbineColour(data["ForegroundColour"]), self:ColourTupleToTurbineColour(data["BackgroundColour"]), data["ShowText"],
                                data["BoxOffset"], data["BorderSize"], self:ColourTupleToTurbineColour(data["BorderColour"]), self:ColourTupleToTurbineColour(data["BoxBackgroundColour"]), data["Full_Border"], data["Animate"],
                                self:ColourTupleToTurbineColour(data["Font_Colour"]), self:FontToFont(data["Font"]))
    
    if self.groupMode then
        resourceBar.moveBox = MoveBox(resourceBar, data["Width"], data["Height"], "Resourcebar")
        resourceBar.savePosition = function()
            data["X"], data["Y"] = resourceBar:GetPosition()
        end
    end
    
    if self.playerClass == Turbine.Gameplay.Class.Champion then
        resourceBar:ChangeResource(self.ClassAttr:GetFervor())
        self.attrCallbacks["FervorChanged"] = AddCallback(self.ClassAttr, "FervorChanged", function(sender, args) resourceBar:ChangeResource( self.ClassAttr:GetFervor() ) end)
    elseif self.playerClass == Turbine.Gameplay.Class.Hunter then
        resourceBar:ChangeResource(self.ClassAttr:GetFocus())
        self.attrCallbacks["FocusChanged"] = AddCallback(self.ClassAttr, "FocusChanged", function(sender, args) resourceBar:ChangeResource( self.ClassAttr:GetFocus() ) end)
    elseif self.playerClass == Turbine.Gameplay.Class.RuneKeeper then
        if data["RKLayout"] then
            local Attunement = self.ClassAttr:GetAttunement()
            if Attunement > 10 then
                resourceBar:ChangeRKResource(math.abs( Attunement - 10), self:ColourTupleToTurbineColour(data["RK_BackgroundColour_Heal"]), self:ColourTupleToTurbineColour(data["RK_ForegroundColour_Heal"]))
            else
                resourceBar:ChangeRKResource(math.abs( Attunement - 10), self:ColourTupleToTurbineColour(data["RK_BackgroundColour_DPS"]), self:ColourTupleToTurbineColour(data["RK_ForegroundColour_DPS"]))
            end
            self.attrCallbacks["AttunementChanged"] = AddCallback(self.ClassAttr, "AttunementChanged", function(sender, args)
                local Attunement = self.ClassAttr:GetAttunement()
                if Attunement > 10 then
                    resourceBar:ChangeRKResource(math.abs( Attunement - 10), self:ColourTupleToTurbineColour(data["RK_BackgroundColour_Heal"]), self:ColourTupleToTurbineColour(data["RK_ForegroundColour_Heal"]))
                else
                    resourceBar:ChangeRKResource(math.abs( Attunement - 10), self:ColourTupleToTurbineColour(data["RK_BackgroundColour_DPS"]), self:ColourTupleToTurbineColour(data["RK_ForegroundColour_DPS"]))
                end         
            end)
            
        else
            resourceBar:ChangeResource(self.ClassAttr:GetAttunement())
            self.attrCallbacks["AttunementChanged"] = AddCallback(self.ClassAttr, "AttunementChanged", function(sender, args) resourceBar:ChangeResource( self.ClassAttr:GetAttunement() ) end)
        end
    elseif self.playerClass == Turbine.Gameplay.Class.Brawler then
        for i = 1, self.player:GetEffects():GetCount(), 1 do
            local effect = self.player:GetEffects():Get(i)
            if string.gsub(effect:GetName(), "%d", "") == "Battle Flow " then
                resourceBar:ChangeResource( tonumber(string.gsub(effect:GetName(), "%D", "")) )
            end
        end

        table.insert(self.effectAddedCallbacks, AddCallback(PlayerEffectsHandler, "EffectAdded", 
        function(effect)
            if string.gsub(effect:GetName(), "%d", "") == "Battle Flow " then
                resourceBar:ChangeResource( tonumber(string.gsub(effect:GetName(), "%D", "")) )
                self.battleFlowID = effect:GetID()
            end
        end))

        table.insert(self.effectRemovedCallbacks, AddCallback(PlayerEffectsHandler, "EffectRemoved", 
        function(effect)
            if self.battleFlowID == effect:GetID() then
                resourceBar:ChangeResource(0)
            end
        end))

    elseif self.playerClass == Turbine.Gameplay.Class.Guardian then
        table.insert(self.effectAddedCallbacks, AddCallback(PlayerEffectsHandler, "EffectAdded", 
        function(effect)
            if self.fortificationTable[effect:GetName()] then
                resourceBar:ChangeResource(self.fortificationTable[effect:GetName()])
                self.forticationID = effect:GetID()
            end
        end))

        table.insert(self.effectRemovedCallbacks, AddCallback(PlayerEffectsHandler, "EffectRemoved", 
        function(effect)
            if self.forticationID == effect:GetID() and self.fortificationTable[effect:GetName()] then
                resourceBar:ChangeResource(0)
            end
        end))
    elseif self.playerClass == Turbine.Gameplay.Class.Beorning then
        for i = 1, self.player:GetEffects():GetCount(), 1 do
            local effect = self.player:GetEffects():Get(i)
            if string.match(effect:GetName(), "Guarded %- Tier (.*)") then
                self.GuardedTierID = effect:GetID()
                resourceBar:ChangeResource(tonumber(string.match(effect:GetName(), "Guarded %- Tier (.*)")))
            end
        end

        table.insert(self.effectAddedCallbacks, AddCallback(PlayerEffectsHandler, "EffectAdded", 
        function(effect)
            if string.match(effect:GetName(), "Guarded %- Tier (.*)") then
                self.GuardedTierID = effect:GetID()
                resourceBar:ChangeResource(tonumber(string.match(effect:GetName(), "Guarded %- Tier (.*)")))
            end
        end))

        table.insert(self.effectRemovedCallbacks, AddCallback(PlayerEffectsHandler, "EffectRemoved", 
        function(effect)
            if self.GuardedTierID == effect:GetID() then
                resourceBar:ChangeResource(0)
            end
        end))
    elseif self.playerClass == Turbine.Gameplay.Class.BlackArrow then
        for i = 1, self.player:GetEffects():GetCount(), 1 do
            local effect = self.player:GetEffects():Get(i)
            if string.find(effect:GetName(), "Keen Eye") then
                self.keenEyeID = effect:GetID()
                if string.match(effect:GetName(), "Keen Eye %((.*)%)") then
                    resourceBar:ChangeResource(tonumber(string.match(effect:GetName(), "Keen Eye %((.*)%)")))
                elseif effect:GetName() == "Keen Eye" then
                    resourceBar:ChangeResource(1)
                end
            end
        end

        table.insert(self.effectAddedCallbacks, AddCallback(PlayerEffectsHandler, "EffectAdded", 
        function(effect)
            if string.find(effect:GetName(), "Keen Eye") then
                self.keenEyeID = effect:GetID()
                if string.match(effect:GetName(), "Keen Eye %((.*)%)") then
                    resourceBar:ChangeResource(tonumber(string.match(effect:GetName(), "Keen Eye %((.*)%)")))
                elseif effect:GetName() == "Keen Eye" then
                    resourceBar:ChangeResource(1)
                end
            end
        end))

        table.insert(self.effectRemovedCallbacks, AddCallback(PlayerEffectsHandler, "EffectRemoved", 
        function(effect)
            if self.keenEyeID == effect:GetID() then
                resourceBar:ChangeResource(0)
            end
        end))
    elseif self.playerClass == Turbine.Gameplay.Class.Weaver then
        for i = 1, self.player:GetEffects():GetCount(), 1 do
            local effect = self.player:GetEffects():Get(i)
            if string.gsub(effect:GetName(), "%d", "") == "Venom " then
                resourceBar:ChangeResource( tonumber(string.gsub(effect:GetName(), "%D", "")) )
            end
        end

        table.insert(self.effectAddedCallbacks, AddCallback(PlayerEffectsHandler, "EffectAdded", 
        function(effect)
            if string.gsub(effect:GetName(), "%d", "") == "Venom " then
                resourceBar:ChangeResource( tonumber(string.gsub(effect:GetName(), "%D", "")) )
                self.venomID = effect:GetID()
            end
        end))

        table.insert(self.effectRemovedCallbacks, AddCallback(PlayerEffectsHandler, "EffectRemoved", 
        function(effect)
            if self.venomID == effect:GetID() then
                resourceBar:ChangeResource(0)
            end
        end))
    end

    table.insert(self.tools, resourceBar)

    self.currentHeight = self.currentHeight + data["Height"]
end

function PrimeUI:SetupMoraleBar(data, parent)
    local valuebar = Turbine.UI.Window()
    valuebar:SetParent(parent)
    if self.groupMode then
        valuebar:SetPosition(data["X"], data["Y"])
        valuebar.moveBox = MoveBox(valuebar, data["Width"] + 2 * data["BorderSize"], data["Height"] + 2 * data["BorderSize"], "Morale Bar")
        valuebar.savePosition = function()
            data["X"], data["Y"] = valuebar:GetPosition()
        end
    else
        valuebar:SetPosition(self:ContentAlignement(data["Alignment"], data["Width"] + 2 * data["BorderSize"]), self.currentHeight)
    end
    valuebar:SetSize(data["Width"] + 2 * data["BorderSize"], data["Height"] + 2 * data["BorderSize"])
    valuebar:SetMouseVisible(false)
    valuebar:SetVisible(true)
    valuebar:SetBackColor(self:ColourTupleToTurbineColour(data["BorderColour"]))

    -- args (parent, x, y, width, height, textFormat, textAlignment, barAlignment, barColours, barbackground)
    local moralebar = PermanentBar(valuebar, data["BorderSize"], data["BorderSize"], data["Width"], data["Height"], data["Text"], Turbine.UI.ContentAlignment.MiddleCenter, data["Alignment"], self:ColourTupleToTurbineColour(data["Colours"]), "bar25", self:FontToFont(data["Font"]), self:ColourTupleToTurbineColour(data["Interval_Colour"]), data["Intervals"])
    moralebar:SetMaxValue(self.player:GetMaxMorale())
    moralebar:SetCurrentValue(self.player:GetMorale())
    moralebar:SetTempValue(self.player:GetTemporaryMorale())

    if data["Smooth"] then
        table.insert(self.playerCallbacks, {[1] = "MoraleChanged", [2] = AddCallback(self.player, "MoraleChanged", function(sender, args) moralebar:SetCurrentValue( self.player:GetMorale()) end)})
    else
        table.insert(self.playerCallbacks, {[1] = "MoraleChanged", [2] = AddCallback(self.player, "MoraleChanged", function(sender, args) moralebar:SetCurrentValueImmediate( self.player:GetMorale()) end)})
    end
    table.insert(self.playerCallbacks, {[1] = "MaxMoraleChanged", [2] = AddCallback(self.player, "MaxMoraleChanged", function(sender, args) moralebar:SetMaxValue( self.player:GetMaxMorale() ) end)})
    table.insert(self.playerCallbacks, {[1] = "BaseMaxMoraleChanged", [2] = AddCallback(self.player, "BaseMaxMoraleChanged", function(sender, args) moralebar:SetMaxValue( self.player:GetMaxMorale() ) end)})
    table.insert(self.playerCallbacks, {[1] = "TemporaryMoraleChanged", [2] = AddCallback(self.player, "TemporaryMoraleChanged", function(sender, args) moralebar:SetTempValue( self.player:GetTemporaryMorale()) end)})
    moralebar:SetCurrentValue(self.player:GetMorale())
    moralebar:SetMaxValue(self.player:GetMaxMorale() )
    table.insert(self.UI, valuebar)
    table.insert(self.tools, moralebar)

    self.currentHeight = self.currentHeight + data["Height"] + 2 * data["BorderSize"]
end

function PrimeUI:SetupTargetMoraleBar(data, parent)
    local valuebar = Turbine.UI.Window()
    valuebar:SetParent(parent)
    if self.groupMode then
        valuebar:SetPosition(data["X"], data["Y"])
        valuebar.moveBox = MoveBox(valuebar, data["Width"] + 2 * data["BorderSize"], data["Height"] + 2 * data["BorderSize"], "Target Morale Bar")
        valuebar.savePosition = function()
            data["X"], data["Y"] = valuebar:GetPosition()
        end
    else
        valuebar:SetPosition(self:ContentAlignement(data["Alignment"], data["Width"] + 2 * data["BorderSize"]), self.currentHeight)
    end
    valuebar:SetSize(data["Width"] + 2 * data["BorderSize"], data["Height"] + 2 * data["BorderSize"])
    valuebar:SetMouseVisible(false)
    valuebar:SetBackColor(self:ColourTupleToTurbineColour(data["BorderColour"]))
    valuebar:SetVisible(false)

    -- args (parent, x, y, width, height, textFormat, textAlignment, barAlignment, barColours, barbackground)
    local moralebar = PermanentBar(valuebar, data["BorderSize"], data["BorderSize"], data["Width"], data["Height"], data["Text"], Turbine.UI.ContentAlignment.MiddleCenter, data["Alignment"], self:ColourTupleToTurbineColour(data["Colours"]), "bar25", self:FontToFont(data["Font"]), self:ColourTupleToTurbineColour(data["Interval_Colour"]), data["Intervals"])
    
    moralebar:SetMaxValue(1)
    moralebar:SetCurrentValue(1)
    moralebar:SetTempValue(0)
    
    if data["Smooth"] then
        table.insert(self.healthTargetCallbacks, TargetHandlerObject:AddHealthCallback(function(value) moralebar:SetCurrentValue(value) end))
    else
        table.insert(self.healthTargetCallbacks, TargetHandlerObject:AddHealthCallback(function(value) moralebar:SetCurrentValueImmediate(value) end))
    end
    table.insert(self.temporaryHealthTargetCallbacks, TargetHandlerObject:AddTemporaryHealthCallback(function(value) moralebar:SetTempValue(value) end))

    table.insert(self.targetInfoCallbacks, TargetHandlerObject:AddTargetInfoCallback(function(value) 
        if value then
            valuebar:SetVisible(true)
            moralebar:SetMaxValue(value["MaxMorale"] or 0)
            moralebar:SetCurrentValueImmediate(value["Morale"] or 0)
            moralebar:SetTempValue(value["TemporaryMorale"] or 0)
        else
            valuebar:SetVisible(false)
        end
    end))

    table.insert(self.UI, valuebar)
    table.insert(self.tools, moralebar)

    self.currentHeight = self.currentHeight + data["Height"] + 2 * data["BorderSize"]
end

function PrimeUI:SetupPowerBar(data, parent)
    local valuebar = Turbine.UI.Control()
    valuebar:SetParent(parent)
    if self.groupMode then
        valuebar:SetPosition(data["X"], data["Y"])
        valuebar.moveBox = MoveBox(valuebar, data["Width"] + 2 * data["BorderSize"], data["Height"] + 2 * data["BorderSize"], "Power Bar")
        valuebar.savePosition = function()
            data["X"], data["Y"] = valuebar:GetPosition()
        end
    else
        valuebar:SetPosition(self:ContentAlignement(data["Alignment"], data["Width"] + 2 * data["BorderSize"]), self.currentHeight)
    end
    valuebar:SetSize(data["Width"] + 2 * data["BorderSize"], data["Height"] + 2 * data["BorderSize"])
    valuebar:SetMouseVisible(false)
    valuebar:SetBackColor(self:ColourTupleToTurbineColour(data["BorderColour"]))

    -- args (parent, x, y, width, height, textFormat, textAlignment, barAlignment, barColours, barbackground)
    local powerbar = PermanentBar(valuebar, data["BorderSize"], data["BorderSize"], data["Width"], data["Height"], data["Text"], Turbine.UI.ContentAlignment.MiddleCenter, data["Alignment"], self:ColourTupleToTurbineColour(data["Colours"]), "bar25", self:FontToFont(data["Font"]), self:ColourTupleToTurbineColour(data["Interval_Colour"]), data["Intervals"])
    powerbar:SetMaxValue(self.player:GetMaxPower())
    powerbar:SetCurrentValue(self.player:GetPower())

    if data["Smooth"] then
        table.insert(self.playerCallbacks, {[1] = "PowerChanged", [2] = AddCallback(self.player, "PowerChanged", function(sender, args) powerbar:SetCurrentValue( self.player:GetPower() ) end)})
    else
        table.insert(self.playerCallbacks, {[1] = "PowerChanged", [2] = AddCallback(self.player, "PowerChanged", function(sender, args) powerbar:SetCurrentValueImmediate( self.player:GetPower() ) end)})
    end
    table.insert(self.playerCallbacks, {[1] = "MaxPowerChanged", [2] = AddCallback(self.player, "MaxPowerChanged", function(sender, args) powerbar:SetMaxValue( self.player:GetMaxPower() ) end)})
    table.insert(self.playerCallbacks, {[1] = "BaseMaxPowerChanged", [2] = AddCallback(self.player, "BaseMaxPowerChanged", function(sender, args) powerbar:SetMaxValue( self.player:GetMaxPower() ) end)})
    table.insert(self.playerCallbacks, {[1] = "TemporaryPowerChanged", [2] = AddCallback(self.player, "TemporaryPowerChanged", function(sender, args) powerbar:SetTempValue( self.player:GetTemporaryPower()) end)})
    powerbar:SetCurrentValue(self.player:GetPower())
    powerbar:SetMaxValue(self.player:GetMaxPower())
    table.insert(self.UI, valuebar)
    table.insert(self.tools, powerbar)

    self.currentHeight = self.currentHeight + data["Height"] + 2 * data["BorderSize"]
end

function PrimeUI:SetupTargetPowerBar(data, parent)
    local valuebar = Turbine.UI.Control()
    valuebar:SetParent(parent)
    if self.groupMode then
        valuebar:SetPosition(data["X"], data["Y"])
        valuebar.moveBox = MoveBox(valuebar, data["Width"] + 2 * data["BorderSize"], data["Height"] + 2 * data["BorderSize"], "Target Power Bar")
        valuebar.savePosition = function()
            data["X"], data["Y"] = valuebar:GetPosition()
        end
    else
        valuebar:SetPosition(self:ContentAlignement(data["Alignment"], data["Width"] + 2 * data["BorderSize"]), self.currentHeight)
    end
    valuebar:SetSize(data["Width"] + 2 * data["BorderSize"], data["Height"] + 2 * data["BorderSize"])
    valuebar:SetMouseVisible(false)
    valuebar:SetBackColor(self:ColourTupleToTurbineColour(data["BorderColour"]))
    valuebar:SetVisible(false)

    -- args (parent, x, y, width, height, textFormat, textAlignment, barAlignment, barColours, barbackground)
    local powerbar = PermanentBar(valuebar, data["BorderSize"], data["BorderSize"], data["Width"], data["Height"], data["Text"], Turbine.UI.ContentAlignment.MiddleCenter, data["Alignment"], self:ColourTupleToTurbineColour(data["Colours"]), "bar25", self:FontToFont(data["Font"]), self:ColourTupleToTurbineColour(data["Interval_Colour"]), data["Intervals"])
    
    powerbar:SetMaxValue(1)
    powerbar:SetCurrentValue(1)
    
    if data["Smooth"] then
        table.insert(self.powerTargetCallbacks, TargetHandlerObject:AddPowerCallback(function(value) powerbar:SetCurrentValue(value) end))
    else
        table.insert(self.powerTargetCallbacks, TargetHandlerObject:AddPowerCallback(function(value) powerbar:SetCurrentValueImmediate(value) end))
    end

    table.insert(self.targetInfoCallbacks, TargetHandlerObject:AddTargetInfoCallback(function(value) 
        if value then
            valuebar:SetVisible(true)
            powerbar:SetCurrentValueImmediate(value["Power"] or 0)
            powerbar:SetCurrentValue(value["MaxPower"] or value["Power"] or 0)
        else
            valuebar:SetVisible(false)
        end
    end))

    table.insert(self.UI, valuebar)
    table.insert(self.tools, powerbar)

    self.currentHeight = self.currentHeight + data["Height"] + 2 * data["BorderSize"]
end

function PrimeUI:SetupWrathBar(data, parent)
    local valuebar = Turbine.UI.Control()
    valuebar:SetParent(parent)
    if self.groupMode then
        valuebar:SetPosition(data["X"], data["Y"])
        valuebar.moveBox = MoveBox(valuebar, data["Width"] + 2 * data["BorderSize"], data["Height"] + 2 * data["BorderSize"], "Wrath Bar")
        valuebar.savePosition = function()
            data["X"], data["Y"] = valuebar:GetPosition()
        end
    else
        valuebar:SetPosition(self:ContentAlignement(data["Alignment"], data["Width"] + 2 * data["BorderSize"]), self.currentHeight)
    end
    valuebar:SetSize(data["Width"] + 2 * data["BorderSize"], data["Height"] + 2 * data["BorderSize"])
    valuebar:SetMouseVisible(false)
    valuebar:SetBackColor(self:ColourTupleToTurbineColour(data["BorderColour"]))
    
    -- args (parent, x, y, width, height, textFormat, textAlignment, barAlignment, barColours, barbackground)
    local wrathBar = PermanentBar(valuebar, data["BorderSize"], data["BorderSize"], data["Width"], data["Height"], data["Text"], Turbine.UI.ContentAlignment.MiddleCenter, data["Alignment"], self:ColourTupleToTurbineColour(data["Colours"]), "bar25", self:FontToFont(data["Font"]), self:ColourTupleToTurbineColour(data["Interval_Colour"]), data["Intervals"])

    if self.playerClass == Turbine.Gameplay.Class.Beorning then
        wrathBar:SetMaxValue(100)
        wrathBar:SetCurrentValue(self.ClassAttr:GetWrath())
        if data["Smooth"] then
            self.attrCallbacks["WrathChanged"] = AddCallback(self.ClassAttr, "WrathChanged", function(sender, args) wrathBar:SetCurrentValue( self.ClassAttr:GetWrath() ) end)
        else
            self.attrCallbacks["WrathChanged"] = AddCallback(self.ClassAttr, "WrathChanged", function(sender, args) wrathBar:SetCurrentValueImmediate( self.ClassAttr:GetWrath() ) end)
        end
        wrathBar:SetCurrentValue( self.ClassAttr:GetWrath() )
    end

    table.insert(self.UI, valuebar)
    table.insert(self.tools, wrathBar)

    self.currentHeight = self.currentHeight + data["Height"] + 2 * data["BorderSize"]
end

function PrimeUI:TimerGroup(data, parent)
    local valuebar = Turbine.UI.Control()
    valuebar:SetParent(parent)
    if self.groupMode then
        valuebar:SetPosition(data["X"], data["Y"])
        valuebar.moveBox = MoveBox(valuebar, data["Width"], data["Height"], "Timer Group")
        valuebar.savePosition = function()
            data["X"], data["Y"] = valuebar:GetPosition()
        end
    else
        valuebar:SetPosition(self:ContentAlignement(data["Alignment"], data["Width"]), self.currentHeight)
    end
    valuebar:SetSize(data["Width"], data["Height"])
    valuebar:SetMouseVisible(false)

    local innerbar = Turbine.UI.Control()
    innerbar:SetParent(valuebar)
    innerbar:SetPosition(data["BorderSize"], data["BorderSize"])
    innerbar:SetSize(data["Width"], data["Height"])
    innerbar:SetMouseVisible(false)
    innerbar:SetBackColor(self:ColourTupleToTurbineColour(data["BackgroundColour"]))

    local currentHeight = 0
    local colourScheme = {self:ColourTupleToTurbineColour(data["Timer_BackgroundColour"]), self:ColourTupleToTurbineColour(data["Timer_BackgroundColour"]), self:ColourTupleToTurbineColour(data["Timer_ForegroundColour"]), Turbine.UI.Color(1,1,1), Turbine.UI.Color(0,0,0), nil, nil, nil, 1, 1}

    local timerGroup = TimerGroupWindow(innerbar, data["Width"], data["Height"], 0, 0, data["Timer_Sorting"], data["Timer_Height"], self:FontToFont(data["Timer_Font"]), self:FontToFont(data["Duration_Font"]), colourScheme, data["Static_Timers"], data["Timer_Offset"], data["Timer_Stack_Direction"], data["Spark"], data["BorderSize"], self:ColourTupleToTurbineColour(data["BorderColour"]))
    timerGroup.targetMaxHealth = 0
    timerGroup.TargetName = ""
    timerGroup.SkillID = {}

    if data["Static_Timers"] then
        if data["EffectAdded"] then
            for key, value in pairs(data["EffectAdded"]) do
                local icon = value["Icon"]
                if value["Icon"] == true then
                    icon = 0x41007F41
                end
                timerGroup:NewTimer(0, string.gsub(value["Message"], "<name>", self.player:GetName()), value["Message"], value["Unique"], icon, self.player:GetName(), true) 
            end
        end

        if data["Effects"] then
            for key, value in pairs(data["Effects"]) do
                local icon = value["Icon"]
                if value["Icon"] == true then
                    icon = 0x41007F41
                end
                timerGroup:NewTimer(0, string.gsub(value["Message"], "<name>", self.player:GetName()), value["Message"], value["Unique"], icon, self.player:GetName(), true) 
            end
        end

        if data["EffectAdded_Target"] then
            for key, value in pairs(data["EffectAdded_Target"]) do
                local icon = value["Icon"]
                if value["Icon"] == true then
                    icon = 0x41007F41
                end
                timerGroup:NewTimer(0, string.gsub(value["Message"], "<name>", self.player:GetName()), value["Message"], value["Unique"], icon, self.player:GetName(), true) 
            end
        end

        if data["EffectAdded_Group"] then
            for key, value in pairs(data["EffectAdded_Group"]) do
                local icon = value["Icon"]
                if value["Icon"] == true then
                    icon = 0x41007F41
                end
                timerGroup:NewTimer(0, string.gsub(value["Message"], "<name>", self.player:GetName()), value["Message"], value["Unique"], icon, self.player:GetName(), true) 
            end
        end

        if data["EffectRemoved"] then
            for key, value in pairs(data["EffectRemoved"]) do
                local icon = value["Icon"]
                if value["Icon"] == true then
                    icon = 0x41007F41
                end
                timerGroup:NewTimer(0, string.gsub(value["Message"], "<name>", self.player:GetName()), value["Message"], value["Unique"], icon, self.player:GetName(), true) 
            end
        end

        if data["Skills"] then
            for key, value in pairs(data["Skills"]) do
                local icon = value["Icon"]
                if value["Icon"] == true then
                    icon = 0x41007F41
                end
                timerGroup:NewTimer(0, string.gsub(value["Message"], "<name>", self.player:GetName()), value["Message"], value["Unique"], icon, self.player:GetName(), true) 
            end
        end

        if data["EnemyHits"] then
            for key, value in pairs(data["EnemyHits"]) do
                timerGroup:NewTimer(0, string.gsub(value["Message"], "<name>", self.player:GetName()), value["Message"], value["Unique"], value["Icon"], self.player:GetName(), true) 
            end
        end

        if data["PlayerHits"] then
            for key, value in pairs(data["PlayerHits"]) do
                timerGroup:NewTimer(0, string.gsub(value["Message"], "<name>", self.player:GetName()), value["Message"], value["Unique"], value["Icon"], self.player:GetName(), true) 
            end
        end
        
        if data["PlayerBenefit"] then
            for key, value in pairs(data["PlayerBenefit"]) do
                timerGroup:NewTimer(0, string.gsub(value["Message"], "<name>", self.player:GetName()), value["Message"], value["Unique"], value["Icon"], self.player:GetName(), true) 
            end
        end
        
        if data["PlayerHeal"] then
            for key, value in pairs(data["PlayerHeal"]) do
                timerGroup:NewTimer(0, string.gsub(value["Message"], "<name>", self.player:GetName()), value["Message"], value["Unique"], value["Icon"], self.player:GetName(), true) 
            end
        end

        if data["StandardChat"] then
            for key, value in pairs(data["StandardChat"]) do
                timerGroup:NewTimer(0, string.gsub(value["Message"], "<name>", self.player:GetName()), value["Message"], value["Unique"], value["Icon"], self.player:GetName(), true) 
            end
        end

        if data["PartyChat"] then
            for key, value in pairs(data["PartyChat"]) do
                timerGroup:NewTimer(0, string.gsub(value["Message"], "<name>", self.player:GetName()), value["Message"], value["Unique"], value["Icon"], self.player:GetName(), true) 
            end
        end

        if data["SayChat"] then
            for key, value in pairs(data["SayChat"]) do
                timerGroup:NewTimer(0, string.gsub(value["Message"], "<name>", self.player:GetName()), value["Message"], value["Unique"], value["Icon"], self.player:GetName(), true) 
            end
        end

        if data["Target_HealthThreshold"] then
            for key, value in pairs(data["Target_HealthThreshold"]) do
                timerGroup:NewTimer(0, string.gsub(value["Message"], "<name>", self.player:GetName()), value["Message"], value["Unique"], value["Icon"], self.player:GetName(), true) 
            end
        end
    end

    if tablelength(data["Effects"]) > 0 then
        timerGroup.effectFunction = function(effect)
            local name = effect:GetName()
            local value = data["Effects"][name]
            if value then
                local time = value["Time"]
                local maxTime = value["Time"]
                if value["Time"] == true then
                    time = effect:GetDuration() - (Turbine.Engine.GetGameTime() - effect:GetStartTime())
                    maxTime = effect:GetDuration()
                end

                local icon = value["Icon"]
                local iconValue = value["Icon"]
                if value["Icon"] == true then
                    icon = effect:GetIcon()
                    iconValue = 0x41007F41
                elseif value["Match_icon"] then
                    if tostring(icon) ~= tostring(effect:GetIcon()) then return end
                end

                if data["Static_Timers"] then
                    timerGroup.SkillID[effect:GetID()] = timerGroup:StartStaticTimer(time, value["Message"], self.player:GetName(), icon, iconValue, data["Unique"], maxTime)
                else
                    timerGroup.SkillID[effect:GetID()] = timerGroup:NewTimer(time, string.gsub(value["Message"], "<name>", self.player:GetName()), value["Message"], value["Unique"], icon, self.player:GetName()) 
                end
            end
        end

        for i = 1, self.player:GetEffects():GetCount() do
            timerGroup.effectFunction(self.player:GetEffects():Get(i))
        end

        table.insert(self.effectAddedCallbacks, AddCallback(PlayerEffectsHandler, "EffectAdded", timerGroup.effectFunction))

        table.insert(self.effectRemovedCallbacks, AddCallback(PlayerEffectsHandler, "EffectRemoved", 
        function(effect)
            local ID = effect:GetID()
            if timerGroup.SkillID[ID] then
                timerGroup:DeleteTimer(timerGroup.SkillID[ID])
            end

        end))
    end

    if tablelength(data["EffectAdded"]) > 0 then
        table.insert(self.effectAddedCallbacks, AddCallback(PlayerEffectsHandler, "EffectAdded", 
        function(effect)
            local name = effect:GetName()
            if data["EffectAdded"][name] then
                local value = data["EffectAdded"][name]
                local time = value["Time"]
                local maxTime = value["Time"]
                if value["Time"] == true then
                    time = effect:GetDuration() - (Turbine.Engine.GetGameTime() - effect:GetStartTime())
                    maxTime = effect:GetDuration()
                end

                local icon = value["Icon"]
                local iconValue = value["Icon"]
                if value["Icon"] == true then
                    icon = effect:GetIcon()
                    iconValue = 0x41007F41
                elseif value["Match_icon"] then
                    if tostring(icon) ~= tostring(effect:GetIcon()) then return end
                end

                if data["Static_Timers"] then
                    timerGroup:StartStaticTimer(time, value["Message"], self.player:GetName(), icon, iconValue, data["Unique"], maxTime)
                else
                    timerGroup:NewTimer(time, string.gsub(value["Message"], "<name>", self.player:GetName()), value["Message"], value["Unique"], icon, self.player:GetName()) 
                end
            end
        end))
    end

    if tablelength(data["EffectAdded_Target"]) > 0 then
        table.insert(self.effectTargetAddedCallbacks, TargetHandlerObject:AddEffectCallback( 
        function(effect)
            local name = effect:GetName()
            if data["EffectAdded_Target"][name] ~= nil then
                local value = data["EffectAdded_Target"][name]
                local time = value["Time"]
                local maxTime = value["Time"]
                if value["Time"] == true then
                    time = effect:GetDuration() - (Turbine.Engine.GetGameTime() - effect:GetStartTime())
                    maxTime = effect:GetDuration()
                end

                local icon = value["Icon"]
                local iconValue = value["Icon"]
                if value["Icon"] == true then
                    icon = effect:GetIcon()
                    iconValue = 0x41007F41
                elseif value["Match_icon"] then
                    if tostring(icon) ~= tostring(effect:GetIcon()) then return end
                end


                if data["Static_Timers"] then
                    timerGroup:StartStaticTimer(time, value["Message"], timerGroup.TargetName, icon, iconValue, data["Unique"], maxTime)
                else
                    timerGroup:NewTimer(time, string.gsub(value["Message"], "<name>", timerGroup.TargetName), value["Message"], value["Unique"], icon, timerGroup.TargetName) 
                end
            end
        end))
    end

    if tablelength(data["EffectAdded_Group"]) > 0 then
        table.insert(self.effectAddedGroupCallbacks, AddCallback(GroupEffectsHandler, "EffectAdded", 
        function(args)
            local effect = args.effect
            local name = effect:GetName()
            if data["EffectAdded_Group"][name] ~= nil then
                local value = data["EffectAdded_Group"][name]
                local time = value["Time"]
                local maxTime = value["Time"]
                if value["Time"] == true then
                    time = effect:GetDuration() - (Turbine.Engine.GetGameTime() - effect:GetStartTime())
                    maxTime = effect:GetDuration()
                end

                local icon = value["Icon"]
                local iconValue = value["Icon"]
                if value["Icon"] == true then
                    icon = effect:GetIcon()
                    iconValue = 0x41007F41
                elseif value["Match_icon"] then
                    if tostring(icon) ~= tostring(effect:GetIcon()) then return end
                end


                if data["Static_Timers"] then
                    timerGroup:StartStaticTimer(time, value["Message"], args.name, icon, iconValue, data["Unique"], maxTime)
                else
                    timerGroup:NewTimer(time, string.gsub(value["Message"], "<name>", args.name), value["Message"], value["Unique"], icon, args.name) 
                end
            end
        end))
    end

    

    if tablelength(data["EffectRemoved"]) > 0 then
        table.insert(self.effectRemovedCallbacks, AddCallback(PlayerEffectsHandler, "EffectRemoved", 
        function(effect)
            local name = effect:GetName()
            if data["EffectRemoved"][name] ~= nil then
                local value = data["EffectRemoved"][name]
                local time = value["Time"]
                local maxTime = value["Time"]
                if value["Time"] == true then
                    time = effect:GetDuration() - (Turbine.Engine.GetGameTime() - effect:GetStartTime())
                    maxTime = effect:GetDuration()
                end

                local icon = value["Icon"]
                local iconValue = value["Icon"]
                if value["Icon"] == true then
                    icon = effect:GetIcon()
                    iconValue = 0x41007F41
                elseif value["Match_icon"] then
                    if tostring(icon) ~= tostring(effect:GetIcon()) then return end
                end

                if data["Static_Timers"] then
                    timerGroup:StartStaticTimer(time, value["Message"], self.player:GetName(), icon, iconValue, data["Unique"], maxTime)
                else
                    timerGroup:NewTimer(time, string.gsub(value["Message"], "<name>", self.player:GetName()), value["Message"], value["Unique"], icon, self.player:GetName()) 
                end
            end
        end))
    end

    if tablelength(data["Skills"]) > 0 then
        self.skills = true
        for key, value in pairs(data["Skills"]) do
            if SkillEventHandler:SkillExists(key) then
                if string.find(value["Message"], "<name>") then
                    timerGroup.skillsTargeted = true
                end
                -- Add the callback for this skill slot
                if not(self.skillCallbacks[key]) then self.skillCallbacks[key] = {} end
                table.insert(self.skillCallbacks[key], AddCallback(SkillEventHandler, key, function(skill)  
                    local time = value["Time"]
                    local maxTime = value["Time"]
                    if value["Time"] == true then
                        time = skill:GetResetTime() - Turbine.Engine.GetGameTime()
                        maxTime = skill:GetCooldown()
                    end

                    local icon = value["Icon"]
                    local iconValue = value["Icon"]
                    if value["Icon"] == true then
                        icon = skill:GetSkillInfo():GetIconImageID()
                        iconValue = 0x41007F41
                    end

                    if data["Static_Timers"] then
                        timerGroup:StartStaticTimer(time, value["Message"], self.player:GetName(), icon, iconValue, data["Unique"], maxTime)
                    else
                        timerGroup:NewTimer(time, string.gsub(value["Message"], "<name>", self.player:GetName()), value["Message"], value["Unique"], icon, self.player:GetName()) 
                    end
                end))
            end
        end
    end

    if tablelength(data["EnemyHits"]) > 0 then
        table.insert(self.enemyChatCallback, AddCallback(ChatEventHandler, "EnemyCombat", function(args)
            local value = data["EnemyHits"][args.SkillName]
            if value then
                local value = value
                if data["Static_Timers"] then
                    timerGroup:StartStaticTimer(value["Time"], string.gsub(value["Message"], "<name>", args.Target), args.Target, value["Icon"], value["Icon"], data["Unique"])
                else
                    timerGroup:NewTimer(value["Time"], string.gsub(value["Message"], "<name>", args.Target), value["Message"], value["Unique"], value["Icon"], args.Target) 
                end
            end
        end))

    end

    if tablelength(data["PlayerHits"]) > 0 then
        table.insert(self.playerChatCallback, AddCallback(ChatEventHandler, "PlayerCombat", function(args)
            if data["PlayerHits"][args.SkillName] then
                local value = data["PlayerHits"][args.SkillName]
                if data["Static_Timers"] then
                    timerGroup:StartStaticTimer(value["Time"], string.gsub(value["Message"], "<name>", args.Target), args.Target, value["Icon"], value["Icon"], data["Unique"])
                else
                    timerGroup:NewTimer(value["Time"], string.gsub(value["Message"], "<name>", args.Target), value["Message"], value["Unique"], value["Icon"], args.Target) 
                end
            end
        end))
    end

    if tablelength(data["PlayerHeal"]) > 0 then
        table.insert(self.playerHealCallbacks, AddCallback(ChatEventHandler, "PlayerHeal", function(args)
            if data["PlayerHeal"][args.SkillName] then
                local value = data["PlayerHeal"][args.SkillName]
                if data["Static_Timers"] then
                    timerGroup:StartStaticTimer(value["Time"], string.gsub(value["Message"], "<name>", args.Target), args.Target, value["Icon"], value["Icon"], data["Unique"])
                else
                    timerGroup:NewTimer(value["Time"], string.gsub(value["Message"], "<name>", args.Target), value["Message"], value["Unique"], value["Icon"], args.Target) 
                end
            end
        end))
    end

    if tablelength(data["PlayerBenefit"]) > 0 then
        table.insert(self.playerBenefitCallbacks, AddCallback(ChatEventHandler, "PlayerBenefit", function(args)
            if data["PlayerBenefit"][args.SkillName] then
                local value = data["PlayerBenefit"][args.SkillName]
                if data["Static_Timers"] then
                    timerGroup:StartStaticTimer(value["Time"], string.gsub(value["Message"], "<name>", args.Target), args.Target, value["Icon"], value["Icon"], data["Unique"])
                else
                    timerGroup:NewTimer(value["Time"], string.gsub(value["Message"], "<name>", args.Target), value["Message"], value["Unique"], value["Icon"], args.Target) 
                end
            end
        end))
    end

    if tablelength(data["StandardChat"]) > 0 then
        table.insert(self.standardChatCallback, AddCallback(ChatEventHandler, "Standard", function(message)
            if data["StandardChat"][message] then
                local value = data["StandardChat"][message]
                if data["Static_Timers"] then
                    timerGroup:StartStaticTimer(value["Time"], value["Message"], self.player:GetName(), value["Icon"], value["Icon"], data["Unique"])
                else
                    timerGroup:NewTimer(value["Time"], string.gsub(value["Message"], "<name>", self.player:GetName()), value["Message"], value["Unique"], value["Icon"], self.player:GetName()) 
                end
            end
        end))
    end

    if tablelength(data["PartyChat"]) > 0 then
        table.insert(self.partyChatCallback, AddCallback(ChatEventHandler, "Party", function(message)
            if data["PartyChat"][message] then
                local value = data["PartyChat"][message]
                if data["Static_Timers"] then
                    timerGroup:StartStaticTimer(value["Time"], value["Message"], self.player:GetName(), value["Icon"], value["Icon"], data["Unique"])
                else
                    timerGroup:NewTimer(value["Time"], string.gsub(value["Message"], "<name>", self.player:GetName()), value["Message"], value["Unique"], value["Icon"], self.player:GetName()) 
                end
            end
        end))
    end

    if tablelength(data["SayChat"]) > 0 then
        table.insert(self.sayChatCallback, AddCallback(ChatEventHandler, "SayLine", function(message)
            if data["SayChat"][message]  then
                local value = data["SayChat"][message]
                if data["Static_Timers"] then
                    timerGroup:StartStaticTimer(value["Time"], value["Message"], self.player:GetName(), value["Icon"], value["Icon"], data["Unique"])
                else
                    timerGroup:NewTimer(value["Time"], string.gsub(value["Message"], "<name>", self.player:GetName()), value["Message"], value["Unique"], value["Icon"], self.player:GetName()) 
                end
            end
        end))
    end

    if tablelength(data["Target_HealthThreshold"]) > 0 or tablelength(data["EffectAdded_Target"]) > 0 or timerGroup.skillsTargeted or data["Reset_On_Retarget"] then
        table.insert(self.targetInfoCallbacks, TargetHandlerObject:AddTargetInfoCallback(function(value) 
            if value then
                if value["Name"] ~= timerGroup.TargetName then
                    timerGroup:Clear()
                end
                timerGroup.targetMaxHealth = value["MaxMorale"]
                timerGroup.TargetName = value["Name"]
            else
                timerGroup.targetMaxHealth = 0
                timerGroup.TargetName = ""
                timerGroup:Clear()
            end
        end))
    end

    
    if tablelength(data["Target_HealthThreshold"]) > 0 then
        timerGroup.targetHealthThresholdCooldowns = {}
        timerGroup.targetHealthThresholdPercentage = {}
        local health = self.player:GetMorale() / self.player:GetMaxMorale() * 100
        for key, value in pairs(data["Target_HealthThreshold"]) do
            timerGroup.targetHealthThresholdCooldowns[key .. tostring(value["Number"]) .. tostring(value["Operator"])] = Turbine.Engine.GetGameTime() - 1000
            timerGroup.targetHealthThresholdPercentage[key .. tostring(value["Number"]) .. tostring(value["Operator"])] = true
        end


        table.insert(self.healthTargetCallbacks, TargetHandlerObject:AddHealthCallback(function(value)
            if timerGroup.targetMaxHealth > 0 then
                local percentage = value / timerGroup.targetMaxHealth * 100
                local time = Turbine.Engine.GetGameTime()
                for key, value in pairs(data["Target_HealthThreshold"]) do
                    if key == timerGroup.TargetName then
                        if self.compareFunctions[value["Operator"]](percentage, value["Number"]) then
                            local id = key .. tostring(value["Number"]) .. tostring(value["Operator"])
                            if timerGroup.targetHealthThresholdCooldowns[id] < time and timerGroup.targetHealthThresholdPercentage[id] then
                                timerGroup.targetHealthThresholdPercentage[id] = false
                                if data["Static_Timers"] then
                                    timerGroup:StartStaticTimer(value["Time"], value["Message"], key, value["Icon"], value["Icon"], data["Unique"])
                                else
                                    timerGroup:NewTimer(value["Time"], string.gsub(value["Message"], "<name>", key), value["Message"], value["Unique"], value["Icon"], key) 
                                end
                                timerGroup.targetHealthThresholdCooldowns[id] = time + 6
                            end
                        else
                            timerGroup.targetHealthThresholdPercentage[key .. tostring(value["Number"]) .. tostring(value["Operator"])] = true
                        end
                    end
                end
            end
        end))
        table.insert(self.combatCallbacks, AddCallback(self.player, "InCombatChanged", function()
            if not(self.player:IsInCombat()) then
                for key, value in pairs(data["Target_HealthThreshold"]) do
                    timerGroup.targetHealthThresholdCooldowns[key .. tostring(value["Number"]) .. tostring(value["Operator"])] = Turbine.Engine.GetGameTime() - 1000
                    timerGroup.targetHealthThresholdPercentage[key .. tostring(value["Number"]) .. tostring(value["Operator"])] = true
                end 
            end
        end))
    end

    table.insert(self.tools, timerGroup)

    self.currentHeight = self.currentHeight + data["Height"] + 2 * data["BorderSize"]

    table.insert(self.UI, valuebar)
end

function PrimeUI:Timer(data, parent)
    local valuebar = Turbine.UI.Control()
    valuebar:SetParent(parent)
    if self.groupMode then
        valuebar:SetPosition(data["X"], data["Y"])
        valuebar.moveBox = MoveBox(valuebar, data["Width"], data["Height"], "Timer")
        valuebar.savePosition = function()
            data["X"], data["Y"] = valuebar:GetPosition()
        end
    else
        valuebar:SetPosition(self:ContentAlignement(data["Alignment"], data["Width"]), self.currentHeight)
    end
    valuebar:SetSize(data["Width"], data["Height"])
    valuebar:SetMouseVisible(false)

    local currentHeight = 0
    local colourScheme = {self:ColourTupleToTurbineColour(data["BackgroundColour"]), self:ColourTupleToTurbineColour(data["BackgroundColour"]), self:ColourTupleToTurbineColour(data["ForegroundColour"]), Turbine.UI.Color(1,1,1), Turbine.UI.Color(0,0,0), nil, nil, nil, 1, 1}

    local timer = EffectTimer(valuebar, data["Width"], data["Height"], nil, data["Text"], self:FontToFont(data["Font"]), colourScheme, true, nil, 0, data["Icon"], nil, not(data["AlwaysVisible"]), self:FontToFont(data["Duration_Font"]), data["Spark"], data["BorderSize"], self:ColourTupleToTurbineColour(data["BorderColour"]))
    timer:SetPosition(data["BorderSize"], data["BorderSize"])
    if not(data["AlwaysVisible"]) then
        timer:SetVisible(false)
    end

    timer.highlight = HighlightFrame2(timer, data["Height"] / 32)
    timer.highlight:Toggle(false) 

    timer.originalMessage = data["Text"]
    timer.currentEffect = ""
    timer.SkillID = {}

    if tablelength(data["Effects"]) > 0 then
        timer.effectFunction = function(effect)
            local name = effect:GetName()
            if data["Effects"][name] then
                if data["Unique"] and timer.currentEffect then
                    if timer.currentEffect == effect:GetName() then
                        return
                    else
                        timer.currentEffect = effect:GetName()
                    end
                end
                local value = data["Effects"][name]
                local time = value["Time"]
                if value["Time"] == true then
                    time = effect:GetDuration()
                end

                local icon = value["Icon"]
                if value["Icon"] == true then
                    icon = effect:GetIcon()
                end

                local highlight = data["AllHighlight"] or value["Highlight"]
                timer.highlight:Toggle(highlight) 
                timer:ChangeIcon(icon)

                timer.SkillID[effect:GetID()] = true
                timer:SetTimer(time, time)
                self.timers[timer] = true
                if not(self.updateCallback) then self.updateCallback = AddCallback(Updater, "ExecuteUpdate", self.ExecuteUpdate) end
            end
        end

        for i = 1, self.player:GetEffects():GetCount() do
            timer.effectFunction(self.player:GetEffects():Get(i))
        end

        table.insert(self.effectAddedCallbacks, AddCallback(PlayerEffectsHandler, "EffectAdded", timer.effectFunction))

        table.insert(self.effectRemovedCallbacks, AddCallback(PlayerEffectsHandler, "EffectRemoved", 
        function(effect)
            local ID = effect:GetID()
            if timer.SkillID[ID] then
                timer:SetTimer(0, 0)
            end

        end))
    end

    if tablelength(data["EffectAdded"]) > 0 then
        table.insert(self.effectAddedCallbacks, AddCallback(PlayerEffectsHandler, "EffectAdded", 
        function(effect)
            local name = effect:GetName()
            if data["EffectAdded"][name] ~= nil then
                if data["Unique"] and timer.currentEffect then
                    if timer.currentEffect == effect:GetName() then
                        return
                    else
                        timer.currentEffect = effect:GetName()
                    end
                end
                local value = data["EffectAdded"][name]
                local time = value["Time"]
                if value["Time"] == true then
                    time = effect:GetDuration()
                end

                local icon = value["Icon"]
                if value["Icon"] == true then
                    icon = effect:GetIcon()
                end

                local highlight = data["AllHighlight"] or value["Highlight"]
                timer.highlight:Toggle(highlight) 
                timer:ChangeIcon(icon)

                timer:SetTimer(time, time)
                self.timers[timer] = true
                if not(self.updateCallback) then self.updateCallback = AddCallback(Updater, "ExecuteUpdate", self.ExecuteUpdate) end
            end
        end))
    end

    if tablelength(data["EffectAdded_Group"]) > 0 then
        table.insert(self.effectAddedGroupCallbacks, AddCallback(GroupEffectsHandler, "EffectAdded", 
        function(args)
            local effect = args.effect
            local name = effect:GetName()
            if data["EffectAdded_Group"][name] ~= nil then
                if data["Unique"] and timer.currentEffect then
                    if timer.currentEffect == effect:GetName() then
                        return
                    else
                        timer.currentEffect = effect:GetName()
                    end
                end
                local value = data["EffectAdded_Group"][name]
                local time = value["Time"]
                if value["Time"] == true then
                    time = effect:GetDuration()
                end

                local icon = value["Icon"]
                if value["Icon"] == true then
                    icon = effect:GetIcon()
                end

                

                local highlight = data["AllHighlight"] or value["Highlight"]

                timer:ChangeText(string.gsub(timer.originalMessage, "<name>", args.name ))
                timer.highlight:Toggle(highlight) 
                timer:ChangeIcon(icon)
                timer:SetTimer(time, time)
                self.timers[timer] = true
                
                if not(self.updateCallback) then self.updateCallback = AddCallback(Updater, "ExecuteUpdate", self.ExecuteUpdate) end
            end
        end))
    end

    if tablelength(data["EffectAdded_Target"]) > 0 then
        table.insert(self.effectTargetAddedCallbacks, TargetHandlerObject:AddEffectCallback( 
        function(effect)
            local name = effect:GetName()
            if data["EffectAdded_Target"][name] ~= nil then
                if data["Unique"] and timer.currentEffect then
                    if timer.currentEffect == effect:GetName() then
                        return
                    else
                        timer.currentEffect = effect:GetName()
                    end
                end
                local value = data["EffectAdded_Target"][name]
                local time = value["Time"]
                if time == true then
                    time = effect:GetDuration()
                end

                local icon = value["Icon"]
                if value["Icon"] == true then
                    icon = effect:GetIcon()
                end

                

                local highlight = data["AllHighlight"] or value["Highlight"]

                timer:ChangeText(string.gsub(timer.originalMessage, "<name>", effect:GetName() ))
                timer.highlight:Toggle(highlight) 
                timer:ChangeIcon(icon)
                timer:SetTimer(time, time)
                self.timers[timer] = true
                
                if not(self.updateCallback) then self.updateCallback = AddCallback(Updater, "ExecuteUpdate", self.ExecuteUpdate) end
            end
        end))
    end

    if tablelength(data["EffectRemoved"]) > 0 then
        table.insert(self.effectRemovedCallbacks, AddCallback(PlayerEffectsHandler, "EffectRemoved", 
        function(effect)
            local name = effect:GetName()
            if data["EffectRemoved"][name] ~= nil then
                if data["Unique"] and timer.currentEffect then
                    if timer.currentEffect == effect:GetName() then
                        return
                    else
                        timer.currentEffect = effect:GetName()
                    end
                end
                local value = data["EffectRemoved"][name]
                local time = value["Time"]
                if time == true then
                    time = effect:GetDuration()
                end

                local icon = value["Icon"]
                if value["Icon"] == true then
                    icon = effect:GetIcon()
                end

                local highlight = data["AllHighlight"] or value["Highlight"]
                timer.highlight:Toggle(highlight) 
                timer:ChangeIcon(icon)

                timer:SetTimer(time, time)
                self.timers[timer] = true
                if not(self.updateCallback) then self.updateCallback = AddCallback(Updater, "ExecuteUpdate", self.ExecuteUpdate) end
            end
        end))
    end

    if tablelength(data["Skills"]) > 0 then
        self.skills = true
        for key, value in pairs(data["Skills"]) do
            if SkillEventHandler:SkillExists(key) then
                -- Add the callback for this skill slot
                if not(self.skillCallbacks[key]) then self.skillCallbacks[key] = {} end
                table.insert(self.skillCallbacks[key], AddCallback(SkillEventHandler, key, function(skill)   
                    if data["Unique"] and timer.currentEffect then
                        if timer.currentEffect == skill:GetSkillInfo():GetName() then
                            return
                        else
                            timer.currentEffect = skill:GetSkillInfo():GetName()
                        end
                    end
                    local time = value["Time"]
                    if value["Time"] == true then
                        time = skill:GetResetTime() - Turbine.Engine.GetGameTime()
                    end

                    local icon = value["Icon"]
                    if value["Icon"] == true then
                        icon = skill:GetSkillInfo():GetIconImageID()
                    end

                    local highlight = data["AllHighlight"] or value["Highlight"]
                    timer.highlight:Toggle(highlight) 
                    timer:ChangeIcon(icon)

                    timer:SetTimer(time, time)
                    if not(self.timers[timer]) then self.timers[timer] = true end
                    if not(self.updateCallback) then self.updateCallback = AddCallback(Updater, "ExecuteUpdate", self.ExecuteUpdate) end
                end))
            end
        end
    end

    if tablelength(data["EnemyHits"]) > 0 then
        table.insert(self.enemyChatCallback, AddCallback(ChatEventHandler, "EnemyCombat", function(args)
            local value = data["EnemyHits"][args.SkillName]
            if value then
                if data["Unique"] and timer.currentEffect then
                    if timer.currentEffect == args.SkillName then
                        return
                    else
                        timer.currentEffect = args.SkillName
                    end
                end
                local highlight = data["AllHighlight"] or value["Highlight"]
                timer.highlight:Toggle(highlight) 
                timer:ChangeIcon(value["Icon"])
                timer:SetTimer(value["Time"], value["Time"])
                self.timers[timer] = true
                if not(self.updateCallback) then self.updateCallback = AddCallback(Updater, "ExecuteUpdate", self.ExecuteUpdate) end
            end
        end))

    end

    if tablelength(data["PlayerHits"]) > 0 then
        table.insert(self.playerChatCallback, AddCallback(ChatEventHandler, "PlayerCombat", function(args)
            local value = data["PlayerHits"][args.SkillName]
            if value then
                if data["Unique"] and timer.currentEffect then
                    if timer.currentEffect == args.SkillName then
                        return
                    else
                        timer.currentEffect = args.SkillName
                    end
                end
                
                local highlight = data["AllHighlight"] or value["Highlight"]
                timer.highlight:Toggle(highlight) 
                timer:ChangeIcon(value["Icon"])
                timer:SetTimer(value["Time"], value["Time"])
                self.timers[timer] = true
                if not(self.updateCallback) then self.updateCallback = AddCallback(Updater, "ExecuteUpdate", self.ExecuteUpdate) end
            end
        end))
    end

    if tablelength(data["PlayerHeal"]) > 0 then
        table.insert(self.playerHealCallbacks, AddCallback(ChatEventHandler, "PlayerHeal", function(args)
            local value = data["PlayerHeal"][args.SkillName]
            if value then
                if data["Unique"] and timer.currentEffect then
                    if timer.currentEffect == args.SkillName then
                        return
                    else
                        timer.currentEffect = args.SkillName
                    end
                end
                
                local highlight = data["AllHighlight"] or value["Highlight"]
                timer.highlight:Toggle(highlight) 
                timer:ChangeIcon(value["Icon"])
                timer:SetTimer(value["Time"], value["Time"])
                self.timers[timer] = true
                if not(self.updateCallback) then self.updateCallback = AddCallback(Updater, "ExecuteUpdate", self.ExecuteUpdate) end
            end
        end))
    end

    if tablelength(data["PlayerBenefit"]) > 0 then
        table.insert(self.playerBenefitCallbacks, AddCallback(ChatEventHandler, "PlayerBenefit", function(args)
            local value = data["PlayerBenefit"][args.SkillName]
            if value then
                if data["Unique"] and timer.currentEffect then
                    if timer.currentEffect == args.SkillName then
                        return
                    else
                        timer.currentEffect = args.SkillName
                    end
                end
                
                local highlight = data["AllHighlight"] or value["Highlight"]
                timer.highlight:Toggle(highlight) 
                timer:ChangeIcon(value["Icon"])
                timer:SetTimer(value["Time"], value["Time"])
                self.timers[timer] = true
                if not(self.updateCallback) then self.updateCallback = AddCallback(Updater, "ExecuteUpdate", self.ExecuteUpdate) end
            end
        end))
    end

    if tablelength(data["StandardChat"]) > 0 then
        table.insert(self.standardChatCallback, AddCallback(ChatEventHandler, "Standard", function(message)
            if data["StandardChat"][message] then
                if data["Unique"] and timer.currentEffect then
                    if timer.currentEffect == message then
                        return
                    else
                        timer.currentEffect = message
                    end
                end
                local value = data["StandardChat"][message]
                local highlight = data["AllHighlight"] or value["Highlight"]
                timer.highlight:Toggle(highlight) 
                timer:ChangeIcon(value["Icon"])
                timer:SetTimer(value["Time"], value["Time"])
                self.timers[timer] = true
                if not(self.updateCallback) then self.updateCallback = AddCallback(Updater, "ExecuteUpdate", self.ExecuteUpdate) end
            end
        end))
    end

    if tablelength(data["PartyChat"]) > 0 then
        table.insert(self.partyChatCallback, AddCallback(ChatEventHandler, "Party", function(message)
            if data["PartyChat"][message] then
                if data["Unique"] and timer.currentEffect then
                    if timer.currentEffect == message then
                        return
                    else
                        timer.currentEffect = message
                    end
                end
                local value = data["PartyChat"][message]
                local highlight = data["AllHighlight"] or value["Highlight"]
                timer.highlight:Toggle(highlight) 
                timer:ChangeIcon(value["Icon"])
                timer:SetTimer(value["Time"], value["Time"])
                self.timers[timer] = true
                if not(self.updateCallback) then self.updateCallback = AddCallback(Updater, "ExecuteUpdate", self.ExecuteUpdate) end
            end
        end))
    end

    if tablelength(data["SayChat"]) > 0 then
        table.insert(self.sayChatCallback, AddCallback(ChatEventHandler, "SayLine", function(message)
            if data["SayChat"][message]  then
                if data["Unique"] and timer.currentEffect then
                    if timer.currentEffect == message then
                        return
                    else
                        timer.currentEffect = message
                    end
                end
                local value = data["SayChat"][message]
                local highlight = data["AllHighlight"] or value["Highlight"]
                timer.highlight:Toggle(highlight) 
                timer:ChangeIcon(value["Icon"])
                timer:SetTimer(value["Time"], value["Time"])
                self.timers[timer] = true
                if not(self.updateCallback) then self.updateCallback = AddCallback(Updater, "ExecuteUpdate", self.ExecuteUpdate) end
            end
        end))
    end

    table.insert(self.tools, timer)

    
    self.currentHeight = self.currentHeight + data["Height"] + 2 * data["BorderSize"]

    table.insert(self.UI, valuebar)
end

function PrimeUI:CircleTimer(data, parent)
    local valuebar = Turbine.UI.Control()
    local width = 132
    local height = 172
    valuebar:SetParent(parent)
    if self.groupMode then
        valuebar:SetPosition(data["X"], data["Y"])
        valuebar.moveBox = MoveBox(valuebar, width, height, "Circle Timer")
        valuebar.savePosition = function()
            data["X"], data["Y"] = valuebar:GetPosition()
        end
    else
        valuebar:SetPosition(self:ContentAlignement(data["Alignment"], width), self.currentHeight)
    end
    valuebar:SetSize(width, height)
    valuebar:SetMouseVisible(false)

    local timer = UI.CircleTimer(valuebar, self:ColourTupleToTurbineColour(data["BackgroundRingColour"]), self:ColourTupleToTurbineColour(data["TimerRingColour"]), self:FontToFont(data["Text_Font"]), self:FontToFont(data["Timer_Font"]), data["Clockwise"], data["Show_Timer"])
    
    timer.targetMaxHealth = 0
    timer.TargetName = ""

    if tablelength(data["EffectAdded"]) > 0 then
        table.insert(self.effectAddedCallbacks, AddCallback(PlayerEffectsHandler, "EffectAdded", 
        function(effect)
            local name = effect:GetName()
            if data["EffectAdded"][name] then
                local value = data["EffectAdded"][name]
                local time = value["Time"]
                if value["Time"] == true then
                    time = effect:GetDuration() - (Turbine.Engine.GetGameTime() - effect:GetStartTime())
                end

                timer:StartTimer(time, string.gsub(value["Message"], "<name>", self.player:GetName()))
            end
        end))
    end

    if tablelength(data["EffectAdded_Target"]) > 0 then
        table.insert(self.effectTargetAddedCallbacks, TargetHandlerObject:AddEffectCallback( 
        function(effect)
            local name = effect:GetName()
            if data["EffectAdded_Target"][name] ~= nil then
                local value = data["EffectAdded_Target"][name]
                local time = value["Time"]
                if value["Time"] == true then
                    time = effect:GetDuration() - (Turbine.Engine.GetGameTime() - effect:GetStartTime())
                end

                timer:StartTimer(time, string.gsub(value["Message"], "<name>", timer.TargetName)) 
            end
        end))
    end

    if tablelength(data["EffectAdded_Group"]) > 0 then
        table.insert(self.effectAddedGroupCallbacks, AddCallback(GroupEffectsHandler, "EffectAdded", 
        function(args)
            local effect = args.effect
            local name = effect:GetName()
            if data["EffectAdded_Group"][name] ~= nil then
                local value = data["EffectAdded_Group"][name]
                local time = value["Time"]
                if value["Time"] == true then
                    time = effect:GetDuration() - (Turbine.Engine.GetGameTime() - effect:GetStartTime())
                end


                timer:StartTimer(time, string.gsub(value["Message"], "<name>", args.name))
            end
        end))
    end

    

    if tablelength(data["EffectRemoved"]) > 0 then
        table.insert(self.effectRemovedCallbacks, AddCallback(PlayerEffectsHandler, "EffectRemoved", 
        function(effect)
            local name = effect:GetName()
            if data["EffectRemoved"][name] ~= nil then
                local value = data["EffectRemoved"][name]
                local time = value["Time"]
                if value["Time"] == true then
                    time = effect:GetDuration() - (Turbine.Engine.GetGameTime() - effect:GetStartTime())
                end

                timer:StartTimer(time, string.gsub(value["Message"], "<name>", self.player:GetName()))
            end
        end))
    end

    if tablelength(data["Skills"]) > 0 then
        self.skills = true
        for key, value in pairs(data["Skills"]) do
            if SkillEventHandler:SkillExists(key) then
                if not(self.skillCallbacks[key]) then self.skillCallbacks[key] = {} end
                table.insert(self.skillCallbacks[key], AddCallback(SkillEventHandler, key, function(skill)  
                    local time = value["Time"]
                    local maxTime = value["Time"]
                    if value["Time"] == true then
                        time = skill:GetResetTime() - Turbine.Engine.GetGameTime()
                    end

                    timer:StartTimer(time, string.gsub(value["Message"], "<name>", self.player:GetName()))
                end))
            end
        end
    end

    if tablelength(data["EnemyHits"]) > 0 then
        table.insert(self.enemyChatCallback, AddCallback(ChatEventHandler, "EnemyCombat", function(args)
            local value = data["EnemyHits"][args.SkillName]
            if value then
                local value = value
                timer:StartTimer(value["Time"], string.gsub(value["Message"], "<name>", args.Target))
            end
        end))

    end

    if tablelength(data["PlayerHits"]) > 0 then
        table.insert(self.playerChatCallback, AddCallback(ChatEventHandler, "PlayerCombat", function(args)
            if data["PlayerHits"][args.SkillName] then
                local value = data["PlayerHits"][args.SkillName]
                timer:StartTimer(value["Time"], string.gsub(value["Message"], "<name>", args.Target))
            end
        end))
    end

    if tablelength(data["PlayerHeal"]) > 0 then
        table.insert(self.playerHealCallbacks, AddCallback(ChatEventHandler, "PlayerHeal", function(args)
            if data["PlayerHeal"][args.SkillName] then
                local value = data["PlayerHeal"][args.SkillName]
                timer:StartTimer(value["Time"], string.gsub(value["Message"], "<name>", args.Target))
            end
        end))
    end

    if tablelength(data["PlayerBenefit"]) > 0 then
        table.insert(self.playerBenefitCallbacks, AddCallback(ChatEventHandler, "PlayerBenefit", function(args)
            if data["PlayerBenefit"][args.SkillName] then
                local value = data["PlayerBenefit"][args.SkillName]
                timer:StartTimer(value["Time"], string.gsub(value["Message"], "<name>", args.Target))
            end
        end))
    end

    if tablelength(data["StandardChat"]) > 0 then
        table.insert(self.standardChatCallback, AddCallback(ChatEventHandler, "Standard", function(message)
            if data["StandardChat"][message] then
                local value = data["StandardChat"][message]
                timer:StartTimer(value["Time"], string.gsub(value["Message"], "<name>", self.player:GetName()))
            end
        end))
    end

    if tablelength(data["PartyChat"]) > 0 then
        table.insert(self.partyChatCallback, AddCallback(ChatEventHandler, "Party", function(message)
            if data["PartyChat"][message] then
                local value = data["PartyChat"][message]
                timer:StartTimer(value["Time"], string.gsub(value["Message"], "<name>", self.player:GetName()))
            end
        end))
    end

    if tablelength(data["SayChat"]) > 0 then
        table.insert(self.sayChatCallback, AddCallback(ChatEventHandler, "SayLine", function(message)
            if data["SayChat"][message]  then
                local value = data["SayChat"][message]
                timer:StartTimer(value["Time"], string.gsub(value["Message"], "<name>", self.player:GetName()))
            end
        end))
    end

    if tablelength(data["Target_HealthThreshold"]) > 0 or tablelength(data["EffectAdded_Target"]) > 0 or data["Reset_On_Retarget"] then
        table.insert(self.targetInfoCallbacks, TargetHandlerObject:AddTargetInfoCallback(function(value) 
            if value then
                if value["Name"] ~= timer.TargetName then
                    timer:Clear()
                end
                timer.targetMaxHealth = value["MaxMorale"]
                timer.TargetName = value["Name"]
            else
                timer.targetMaxHealth = 0
                timer.TargetName = ""
                timer:Clear()
            end
        end))
    end

    
    if tablelength(data["Target_HealthThreshold"]) > 0 then
        timer.targetHealthThresholdCooldowns = {}
        timer.targetHealthThresholdPercentage = {}
        local health = self.player:GetMorale() / self.player:GetMaxMorale() * 100
        for key, value in pairs(data["Target_HealthThreshold"]) do
            timer.targetHealthThresholdCooldowns[key .. tostring(value["Number"]) .. tostring(value["Operator"])] = Turbine.Engine.GetGameTime() - 1000
            timer.targetHealthThresholdPercentage[key .. tostring(value["Number"]) .. tostring(value["Operator"])] = true
        end


        table.insert(self.healthTargetCallbacks, TargetHandlerObject:AddHealthCallback(function(value)
            if timer.targetMaxHealth > 0 then
                local percentage = value / timer.targetMaxHealth * 100
                local time = Turbine.Engine.GetGameTime()
                for key, value in pairs(data["Target_HealthThreshold"]) do
                    if key == timer.TargetName then
                        if self.compareFunctions[value["Operator"]](percentage, value["Number"]) then
                            local id = key .. tostring(value["Number"]) .. tostring(value["Operator"])
                            if timer.targetHealthThresholdCooldowns[id] < time and timer.targetHealthThresholdPercentage[id] then
                                timer.targetHealthThresholdPercentage[id] = false
                                timer:StartTimer(value["Time"], string.gsub(value["Message"], "<name>", key)) 
                                timer.targetHealthThresholdCooldowns[id] = time + 6
                            end
                        else
                            timer.targetHealthThresholdPercentage[key .. tostring(value["Number"]) .. tostring(value["Operator"])] = true
                        end
                    end
                end
            end
        end))
        table.insert(self.combatCallbacks, AddCallback(self.player, "InCombatChanged", function()
            if not(self.player:IsInCombat()) then
                for key, value in pairs(data["Target_HealthThreshold"]) do
                    timer.targetHealthThresholdCooldowns[key .. tostring(value["Number"]) .. tostring(value["Operator"])] = Turbine.Engine.GetGameTime() - 1000
                    timer.targetHealthThresholdPercentage[key .. tostring(value["Number"]) .. tostring(value["Operator"])] = true
                end 
            end
        end))
    end

    table.insert(self.tools, timer)

    self.currentHeight = self.currentHeight + height

    table.insert(self.UI, valuebar)
end

function PrimeUI:Icon(data, parent)
    local valuebar = Turbine.UI.Window()
    valuebar:SetParent(parent)
    if self.groupMode then
        valuebar:SetPosition(data["X"], data["Y"])
        valuebar.moveBox = MoveBox(valuebar, (data["Width"] + 2 * data["BorderSize"]) * data["Scale"], (data["Height"] + 2 * data["BorderSize"]) * data["Scale"], "Icon")
        valuebar.savePosition = function()
            data["X"], data["Y"] = valuebar:GetPosition()
        end
    else
        valuebar:SetPosition(self:ContentAlignement(data["Alignment"], (data["Width"] + 2 * data["BorderSize"]) * data["Scale"]), self.currentHeight)
    end
    valuebar:SetSize((data["Width"] + 2 * data["BorderSize"]) * data["Scale"], (data["Height"] + 2 * data["BorderSize"]) * data["Scale"])
    --valuebar:SetSize(data["Width"] + 2 * data["BorderSize"], data["Height"] + 2 * data["BorderSize"])
    valuebar:SetMouseVisible(false)
    valuebar:SetBackColor(self:ColourTupleToTurbineColour(data["BorderColour"]))

    valuebar.innerbar = Turbine.UI.Control()
    valuebar.innerbar:SetParent(valuebar)
    valuebar.innerbar:SetPosition(data["BorderSize"] * data["Scale"], data["BorderSize"]  * data["Scale"])
    valuebar.innerbar:SetBackground(BackgroundHandler:GetBackground(tonumber(data["Icon"]) or data["Icon"]))
    valuebar.innerbar:SetSize(data["Width"], data["Height"])
    valuebar.innerbar:SetMouseVisible(false)
    valuebar.innerbar:SetBlendMode(Turbine.UI.BlendMode.AlphaBlend)

    if data["Text"] and string.len(data["Text"]) > 0 then
        valuebar.textbar = Turbine.UI.Label()
        valuebar.textbar:SetParent(valuebar)
        valuebar.textbar:SetPosition(data["BorderSize"] * data["Scale"], data["BorderSize"]  * data["Scale"])
        valuebar.textbar:SetSize(data["Width"], data["Height"])
        valuebar.textbar:SetMouseVisible(false)
        valuebar.textbar:SetBlendMode(Turbine.UI.BlendMode.AlphaBlend)
        valuebar.textbar:SetFont(self:FontToFont(data["Font"]))
        valuebar.textbar:SetForeColor(self:ColourTupleToTurbineColour(data["Font_Colour"]))
        valuebar.textbar:SetTextAlignment(self.alignments[data["TextAlignment"]])
        valuebar.textbar:SetText(data["Text"])
    end

    valuebar.innerbar:SetStretchMode(1)
    valuebar.innerbar:SetSize((data["Width"]) * data["Scale"], (data["Height"]) * data["Scale"])

    valuebar.statusTable = {}
    for key, value in pairs(data["Effects"]) do
        if value == false then
            valuebar.statusTable[key] = 1
        else
            valuebar.statusTable[key] = 0
        end
    end

    valuebar.togglefunction = function()
        local tableValue = 0
        for key, value in pairs(valuebar.statusTable) do
            if value > 0 then
                tableValue = tableValue + 1
            end
        end
        local visibilityValue = tableValue >= valuebar.minCount
        valuebar:SetVisible(visibilityValue)
        if valuebar.innerbar.highlight then valuebar.innerbar.highlight:Toggle(visibilityValue) end
    end

    valuebar.count = 0
    if data["Conjuction"] then
        valuebar.minCount = tablelength(data["Effects"]) + tablelength(data["HealthThreshold"])
    else
        valuebar.minCount = 1
    end

    if data["Highlight"] then
        valuebar.innerbar.highlight = HighlightFrame2(valuebar.innerbar, data["Scale"])
        valuebar.innerbar.highlight:Toggle(true) 
    end

    if data["AlwaysVisible"] then
        valuebar:SetVisible(true)
        if valuebar.innerbar.highlight then valuebar.innerbar.highlight:Toggle(true) end
    else
        for i = 1, self.player:GetEffects():GetCount(), 1 do
            local effect = self.player:GetEffects():Get(i)
            if data["Effects"][effect:GetName()] == true then
                valuebar.statusTable[effect:GetName()] = valuebar.statusTable[effect:GetName()] + 1
            elseif data["Effects"][effect:GetName()] == false then
                valuebar.statusTable[effect:GetName()] = valuebar.statusTable[effect:GetName()] - 1
            end
        end


    
        if tablelength(data["Effects"]) then
            table.insert(self.effectAddedCallbacks, AddCallback(PlayerEffectsHandler, "EffectAdded", 
            function(effect)
                local name = effect:GetName()
                if data["Effects"][name] ~= nil then
                    local value = data["Effects"][name]
                    if value == true then
                        valuebar.statusTable[name] = valuebar.statusTable[name] + 1
                    else
                        valuebar.statusTable[name] = valuebar.statusTable[name] - 1
                    end
                    valuebar.togglefunction()
                end
            end))
    
            table.insert(self.effectRemovedCallbacks, AddCallback(PlayerEffectsHandler, "EffectRemoved", 
            function(effect)
                local name = effect:GetName()
                if data["Effects"][name] ~= nil then
                    local value = data["Effects"][effect:GetName()]
                    if value == true then
                        valuebar.statusTable[name] = valuebar.statusTable[name] - 1
                    else
                        valuebar.statusTable[name] = valuebar.statusTable[name] + 1
                    end

                    valuebar.togglefunction()
                end
            end))
        end

        if tablelength(data["HealthThreshold"]) > 0 then
            local health = self.player:GetMorale() / self.player:GetMaxMorale() * 100
            valuebar.healthThresholdPercentage = {}
            for key, value in pairs(data["HealthThreshold"]) do
                local checkValue = (self.compareFunctions[value["Operator"]](health, key) and 1) or 0
                valuebar.statusTable[tostring(key) .. tostring(value["Operator"])] = checkValue
                valuebar.healthThresholdPercentage[tostring(key) .. tostring(value["Operator"])] = checkValue
            end
            table.insert(self.playerHealthCallbacks, AddCallback(self.player, "MoraleChanged", function()
                local percentage = self.player:GetMorale() / self.player:GetMaxMorale() * 100
                for key, value in pairs(data["HealthThreshold"]) do
                    local id = tostring(key) .. tostring(value["Operator"])
                    if self.compareFunctions[value["Operator"]](percentage, key) then
                        if valuebar.healthThresholdPercentage[id] == 0 then
                            if valuebar.healthThresholdPercentage[id] == 0 then
                                valuebar.healthThresholdPercentage[id] = 1
                                valuebar.statusTable[tostring(key) .. tostring(value["Operator"])] = valuebar.statusTable[tostring(key) .. tostring(value["Operator"])] + 1
                            end
                        end
                    else
                        if valuebar.healthThresholdPercentage[id] == 1 then
                            valuebar.healthThresholdPercentage[id] = 0
                            valuebar.statusTable[tostring(key) .. tostring(value["Operator"])] = valuebar.statusTable[tostring(key) .. tostring(value["Operator"])] - 1
                        end
                    end
                end
                valuebar.togglefunction()        
            end))
        end

        valuebar.togglefunction()
    end

    self.currentHeight = self.currentHeight +  ((data["Height"] + 2 * data["BorderSize"]) * data["Scale"])

    if valuebar.textbar then table.insert(self.UI, valuebar.textLabel) end
    table.insert(self.UI, valuebar.innerbar)
    table.insert(self.UI, valuebar)
end

function PrimeUI:TimerIcon(data, parent)
    local valuebar = Turbine.UI.Control()
    valuebar:SetParent(parent)
    if self.groupMode then
        valuebar:SetPosition(data["X"], data["Y"])
        valuebar.moveBox = MoveBox(valuebar, data["Width"] + 2 * data["BorderSize"], data["Height"] + 2 * data["BorderSize"], "Timer Icon")
        valuebar.savePosition = function()
            data["X"], data["Y"] = valuebar:GetPosition()
        end
    else
        valuebar:SetPosition(self:ContentAlignement(data["Alignment"], (data["Width"] + 2 * data["BorderSize"]) * data["Scale"]), self.currentHeight)
    end
    valuebar:SetSize(data["Width"] + 2 * data["BorderSize"], data["Height"] + 2 * data["BorderSize"])
    valuebar:SetMouseVisible(false)
    valuebar:SetBackColor(self:ColourTupleToTurbineColour(data["BorderColour"]))

    local innerbar = BleedTimerLabel(valuebar, data["BorderSize"], data["BorderSize"], 32, 32 * data["Scale"], data["Icon"], true, data["Clockwise"], 0, 1, 1, data["Hide_While_Not_Active"], data["Visual_Timer"], data["ShowText"], data["GrayOut"], data["Hide_While_Not_Active"], data["Visibility_Threshold"], self:FontToFont(data["Font"]))

    innerbar:SetVisible(not(data["Hide_While_Not_Active"]))
    --valuebar:SetStretchMode(1)
    innerbar:SetPosition(data["BorderSize"] * data["Scale"], data["BorderSize"] * data["Scale"])
    valuebar:SetSize((data["Width"] + 2 * data["BorderSize"]) * data["Scale"], (data["Height"] + 2 * data["BorderSize"]) * data["Scale"])

    --valuebar:SetVisible(false)
    if innerbar.highlight then innerbar.highlight:Toggle(false) end
    innerbar.highlightWhileOff = data["HighlightWhileOff"]

    if data["Highlight"] then
        innerbar.highlight = HighlightFrame2(innerbar, data["Scale"])
        innerbar.highlight:Toggle(data["HighlightWhileOff"])
    end

    if tablelength(data["EffectAdded"]) > 0 then
        table.insert(self.effectAddedCallbacks, AddCallback(PlayerEffectsHandler, "EffectAdded", 
        function(effect)
            local name = effect:GetName()
            if data["EffectAdded"][name] ~= nil then
                local time = data["EffectAdded"][name]
                local maxTime = data["EffectAdded"][name]
                if time == true then
                    time = effect:GetDuration() - (Turbine.Engine.GetGameTime() - effect:GetStartTime())
                    maxTime = effect:GetDuration()
                end
                innerbar:SetTimer(time, maxTime)
                if innerbar.highlight then innerbar.highlight:Toggle(not(data["HighlightWhileOff"])) end
                self.icons[innerbar] = true
                if not(self.updateCallback) then self.updateCallback = AddCallback(Updater, "ExecuteUpdate", self.ExecuteUpdate) end
            end
        end))
    end

    if tablelength(data["EffectAdded_Target"]) > 0 then
        table.insert(self.effectTargetAddedCallbacks, TargetHandlerObject:AddEffectCallback( 
        function(effect)
            local name = effect:GetName()
            if data["EffectAdded_Target"][name] ~= nil then
                local time = data["EffectAdded_Target"][name]
                local maxTime = data["EffectAdded_Target"][name]
                if time == true then
                    time = effect:GetDuration() - (Turbine.Engine.GetGameTime() - effect:GetStartTime())
                    maxTime = effect:GetDuration()
                end
                innerbar:SetTimer(time, maxTime)
                if innerbar.highlight then innerbar.highlight:Toggle(not(data["HighlightWhileOff"])) end
                self.icons[innerbar] = true
                if not(self.updateCallback) then self.updateCallback = AddCallback(Updater, "ExecuteUpdate", self.ExecuteUpdate) end
            end
        end))
    end

    if tablelength(data["EffectAdded_Group"]) > 0 then
        table.insert(self.effectAddedGroupCallbacks, AddCallback(GroupEffectsHandler, "EffectAdded", 
        function(args)
            local effect = args.effect
            local name = effect:GetName()
            if data["EffectAdded_Group"][name] ~= nil then
                local time = data["EffectAdded_Group"][name]
                local maxTime = data["EffectAdded_Group"][name]
                if time == true then
                    time = effect:GetDuration() - (Turbine.Engine.GetGameTime() - effect:GetStartTime())
                    maxTime = effect:GetDuration()
                end
                innerbar:SetTimer(time, maxTime)
                if innerbar.highlight then innerbar.highlight:Toggle(not(data["HighlightWhileOff"])) end
                self.icons[innerbar] = true
                if not(self.updateCallback) then self.updateCallback = AddCallback(Updater, "ExecuteUpdate", self.ExecuteUpdate) end
            end
        end))
    end

    if tablelength(data["EffectRemoved"]) > 0 then
        table.insert(self.effectRemovedCallbacks, AddCallback(PlayerEffectsHandler, "EffectRemoved", 
        function(effect)
            local name = effect:GetName()
            if data["EffectRemoved"][name] ~= nil then
                local time = data["EffectRemoved"][name]
                local maxTime = data["EffectRemoved"][name]
                if time == true then
                    time = effect:GetDuration() - (Turbine.Engine.GetGameTime() - effect:GetStartTime())
                    maxTime = effect:GetDuration()
                end
                innerbar:SetTimer(time, maxTime)
                if innerbar.highlight then innerbar.highlight:Toggle(not(data["HighlightWhileOff"])) end
                self.icons[innerbar] = true
                if not(self.updateCallback) then self.updateCallback = AddCallback(Updater, "ExecuteUpdate", self.ExecuteUpdate) end
            end
        end))
    end

    if tablelength(data["EnemyHits"]) > 0 then
        table.insert(self.enemyChatCallback, AddCallback(ChatEventHandler, "EnemyCombat", function(args)
            local value = data["EnemyHits"][args.SkillName]
            if value then
                innerbar:SetTimer(value, value)
                if innerbar.highlight then innerbar.highlight:Toggle(not(data["HighlightWhileOff"])) end
                self.icons[innerbar] = true
                if not(self.updateCallback) then self.updateCallback = AddCallback(Updater, "ExecuteUpdate", self.ExecuteUpdate) end
            end
        end))

    end

    if tablelength(data["PlayerHits"]) > 0 then
        table.insert(self.playerChatCallback, AddCallback(ChatEventHandler, "PlayerCombat", function(args)
            local value = data["PlayerHits"][args.SkillName]
            if value then
                innerbar:SetTimer(value, value)
                if innerbar.highlight then innerbar.highlight:Toggle(not(data["HighlightWhileOff"])) end
                self.icons[innerbar] = true
                if not(self.updateCallback) then self.updateCallback = AddCallback(Updater, "ExecuteUpdate", self.ExecuteUpdate) end
            end
        end))
    end

    if tablelength(data["StandardChat"]) > 0 then
        table.insert(self.standardChatCallback, AddCallback(ChatEventHandler, "Standard", function(message)
            if data["StandardChat"][message] then
                innerbar:SetTimer(data["StandardChat"][message], data["StandardChat"][message])
                if innerbar.highlight then innerbar.highlight:Toggle(not(data["HighlightWhileOff"])) end
                self.icons[innerbar] = true
                if not(self.updateCallback) then self.updateCallback = AddCallback(Updater, "ExecuteUpdate", self.ExecuteUpdate) end
            end
        end))
    end

    if tablelength(data["PartyChat"]) > 0 then
        table.insert(self.partyChatCallback, AddCallback(ChatEventHandler, "Party", function(message)
            if data["PartyChat"][message] then
                innerbar:SetTimer(data["PartyChat"][message], data["PartyChat"][message])
                if innerbar.highlight then innerbar.highlight:Toggle(not(data["HighlightWhileOff"])) end
                self.icons[innerbar] = true
                if not(self.updateCallback) then self.updateCallback = AddCallback(Updater, "ExecuteUpdate", self.ExecuteUpdate) end
            end
        end))
    end

    if tablelength(data["SayChat"]) > 0 then
        table.insert(self.sayChatCallback, AddCallback(ChatEventHandler, "SayLine", function(message)
            if data["SayChat"][message]  then
                innerbar:SetTimer(data["SayChat"][message], data["SayChat"][message])
                if innerbar.highlight then innerbar.highlight:Toggle(not(data["HighlightWhileOff"])) end
                self.icons[innerbar] = true
                if not(self.updateCallback) then self.updateCallback = AddCallback(Updater, "ExecuteUpdate", self.ExecuteUpdate) end
            end
        end))
    end

    if tablelength(data["PlayerHeal"]) > 0 then
        table.insert(self.playerHealCallbacks, AddCallback(ChatEventHandler, "PlayerHeal", function(args)
            if data["PlayerHeal"][args.SkillName]  then
                innerbar:SetTimer(data["PlayerHeal"][args.SkillName], data["PlayerHeal"][args.SkillName])
                if innerbar.highlight then innerbar.highlight:Toggle(not(data["HighlightWhileOff"])) end
                self.icons[innerbar] = true
                if not(self.updateCallback) then self.updateCallback = AddCallback(Updater, "ExecuteUpdate", self.ExecuteUpdate) end
            end
        end))
    end

    if tablelength(data["PlayerBenefit"]) > 0 then
        table.insert(self.playerBenefitCallbacks, AddCallback(ChatEventHandler, "PlayerBenefit", function(args)
            if data["PlayerBenefit"][args.SkillName]  then
                innerbar:SetTimer(data["PlayerBenefit"][args.SkillName], data["PlayerBenefit"][args.SkillName])
                if innerbar.highlight then innerbar.highlight:Toggle(not(data["HighlightWhileOff"])) end
                self.icons[innerbar] = true
                if not(self.updateCallback) then self.updateCallback = AddCallback(Updater, "ExecuteUpdate", self.ExecuteUpdate) end
            end
        end))
    end


    if tablelength(data["Skills"]) > 0 then
        self.skills = true
        for key, value in pairs(data["Skills"]) do
            if SkillEventHandler:SkillExists(key) then
                -- Add the callback for this skill slot
                if not(self.skillCallbacks[key]) then self.skillCallbacks[key] = {} end
                table.insert(self.skillCallbacks[key], AddCallback(SkillEventHandler, key, function(skill) 
                    if innerbar.highlight then innerbar.highlight:Toggle(not(data["HighlightWhileOff"])) end
                    local time = data["Skills"][key]
                    local maxTime = data["Skills"][key]
                    if data["Skills"][key] == true then
                        time = skill:GetResetTime() - Turbine.Engine.GetGameTime()
                        maxTime = skill:GetCooldown()
                    end
                    innerbar:SetTimer(time, maxTime)
                    self.icons[innerbar] = true
                    if not(self.updateCallback) then self.updateCallback = AddCallback(Updater, "ExecuteUpdate", self.ExecuteUpdate) end
                end))
            end
        end
    end
    
    self.currentHeight = self.currentHeight + (data["Height"] + data["BorderSize"] * 2) * data["Scale"]

    table.insert(self.UI, valuebar)
    table.insert(self.tools, innerbar)
end

function PrimeUI:AlertGroup(data, parent)
    local x = data.X
    local y = data.Y
    if not(self.groupMode) then
        x = self:ContentAlignement(data["Alignment"], (data["Width"] + 2 * data["BorderSize"]) * data["Scale"])
        y = self.currentHeight
    end

    local valuebar = AlertGroupPrimeUI(parent, x, y, data.Width, data.Height, data.BorderSize, data.Scale, self:ColourTupleToTurbineColour(data.BorderColour), self:ColourTupleToTurbineColour(data.BackgroundColour), self:ColourTupleToTurbineColour(data.AlertBackgroundColour), self:ColourTupleToTurbineColour(data.AlertBorderColour), self:ColourTupleToTurbineColour(data.Font_Colour), self:FontToFont(data.Font), Turbine.UI.ContentAlignment[data.TextAlignment], self:FontToFont(data.Duration_Font), data.Show_Duration, Turbine.UI.ContentAlignment[data.DurationTextAlignment], data.Decimal_Threshold, self:ColourTupleToTurbineColour(data.Outline_Colour), data.AlertHeight, data.AlertBorderSize, data.ReverseFill)
    --function AlertGroupPrimeUI:Constructor(parent, x, y, width, height, borderSize, scale, borderColour, backgroundColour, alertBackgroundColour, alertBorderColour, fontColour, font, alignment, durationFont, showDuration, durationAlignment, decimalThreshold, outlineColour, alertHeight)
    if self.groupMode then
        valuebar.moveBox = MoveBox(valuebar, data["Width"], data["Height"], "AlertGroup")
        valuebar.savePosition = function()
            if valuebar.moveBox.MoveBox:IsVisible() then
                valuebar:SetVisible(true)
            else
                valuebar:SetVisible(false)
            end
            data["X"], data["Y"] = valuebar:GetPosition()
        end
    end

    if tablelength(data["EffectAdded"]) > 0 then
        table.insert(self.effectAddedCallbacks, AddCallback(PlayerEffectsHandler, "EffectAdded", 
        function(effect)
            local value = data["EffectAdded"][effect:GetName()]
            if value then
                local time = value.Countdown
                if time == true then
                    time = effect:GetDuration() - (Turbine.Engine.GetGameTime() - effect:GetStartTime())
                end

                local icon
                if value["Show_Icon"] then
                    icon = value["Icon"]
                    if value["Icon"] == true then
                        icon = effect:GetIcon()
                    end
                end

                if value["Match_Icon"] then
                    if tostring(icon) ~= tostring(effect:GetIcon()) then return end
                end

                valuebar:AddAlert(icon, time, value.Text)
            end
        end))
    end

    if tablelength(data["EffectAdded_Target"]) > 0 then
        table.insert(self.effectTargetAddedCallbacks, TargetHandlerObject:AddEffectCallback( 
        function(effect)
            local value = data["EffectAdded_Target"][effect:GetName()]
            if value then
                local time = value.Countdown
                if time == true then
                    time = effect:GetDuration() - (Turbine.Engine.GetGameTime() - effect:GetStartTime())
                end

                local icon
                if value["Show_Icon"] then
                    icon = value["Icon"]
                    if value["Icon"] == true then
                        icon = effect:GetIcon()
                    end
                end
                
                if value["Match_Icon"] then
                    if tostring(icon) ~= tostring(effect:GetIcon()) then return end
                end

                valuebar:AddAlert(icon, time, value.Text)
            end
        end))
    end

    if tablelength(data["EffectAdded_Group"]) > 0 then
        table.insert(self.effectAddedGroupCallbacks, AddCallback(GroupEffectsHandler, "EffectAdded", 
        function(args)
            local value = data["EffectAdded_Group"][args.effect:GetName()]
            if value then
                local time = value.Countdown
                if time == true then
                    time = args.effect:GetDuration() - (Turbine.Engine.GetGameTime() - args.effect:GetStartTime())
                end

                local icon
                if value["Show_Icon"] then
                    icon = value["Icon"]
                    if value["Icon"] == true then
                        icon = args.effect:GetIcon()
                    end
                end
                
                if value["Match_Icon"] then
                    if tostring(icon) ~= tostring(args.effect:GetIcon()) then return end
                end

                valuebar:AddAlert(icon, time, string.gsub(value.Text, "<name>", args.name ))
            end
        end))
    end

    if tablelength(data["EffectRemoved"]) > 0 then
        table.insert(self.effectRemovedCallbacks, AddCallback(PlayerEffectsHandler, "EffectRemoved", 
        function(effect)
            local value = data["EffectRemoved"][effect:GetName()]
            if value then
                local time = value.Countdown
                if time == true then
                    time = effect:GetDuration() - (Turbine.Engine.GetGameTime() - effect:GetStartTime())
                end

                local icon
                if value["Show_Icon"] then
                    icon = value["Icon"]
                    if value["Icon"] == true then
                        icon = effect:GetIcon()
                    end
                end
                
                if value["Match_Icon"] then
                    if tostring(icon) ~= tostring(effect:GetIcon()) then return end
                end

                valuebar:AddAlert(icon, time, value.Text)
            end
        end))
    end

    if tablelength(data["Skills"]) > 0 then
        self.skills = true
        for key, value in pairs(data["Skills"]) do
            if SkillEventHandler:SkillExists(key) then
                -- Add the callback for this skill slot
                if not(self.skillCallbacks[key]) then self.skillCallbacks[key] = {} end
                table.insert(self.skillCallbacks[key], AddCallback(SkillEventHandler, key, function(skill) 
                    local time = value.Countdown
                    if time == true then
                        time = skill:GetResetTime() - Turbine.Engine.GetGameTime()
                    end

                    valuebar:AddAlert(value["Icon"], time, value.Text)
                end))
            end
        end
    end

    if tablelength(data["EnemyHits"]) > 0 then
        table.insert(self.enemyChatCallback, AddCallback(ChatEventHandler, "EnemyCombat", function(args)
            local value = data["EnemyHits"][args.SkillName]
            if value then
                valuebar:AddAlert(value["Icon"], value.Countdown, value.Text)
            end
        end))

    end

    if tablelength(data["PlayerHits"]) > 0 then
        table.insert(self.playerChatCallback, AddCallback(ChatEventHandler, "PlayerCombat", function(args)
            local value = data["PlayerHits"][args.SkillName]
            if value then
                valuebar:AddAlert(value["Icon"], value.Countdown, value.Text)
            end
        end))
    end

    if tablelength(data["StandardChat"]) > 0 then
        table.insert(self.standardChatCallback, AddCallback(ChatEventHandler, "Standard", function(message)
            local value = data["StandardChat"][message]
            if value then
                valuebar:AddAlert(value["Icon"], value.Countdown, value.Text)
            end
        end))
    end

    if tablelength(data["PartyChat"]) > 0 then
        table.insert(self.partyChatCallback, AddCallback(ChatEventHandler, "Party", function(message)
            local value = data["PartyChat"][message]
            if value then
                valuebar:AddAlert(value["Icon"], value.Countdown, value.Text)
            end
        end))
    end

    if tablelength(data["SayChat"]) > 0 then
        table.insert(self.sayChatCallback, AddCallback(ChatEventHandler, "SayLine", function(message)
            local value = data["SayChat"][message]
            if value then
                valuebar:AddAlert(value["Icon"], value.Countdown, value.Text)
            end
        end))
    end

    if tablelength(data["PlayerHeal"]) > 0 then
        table.insert(self.playerHealCallbacks, AddCallback(ChatEventHandler, "PlayerHeal", function(args)
            local value = data["PlayerHeal"][args.SkillName]
            if value then
                valuebar:AddAlert(value["Icon"], value.Countdown, value.Text)
            end
        end))
    end

    if tablelength(data["PlayerBenefit"]) > 0 then
        table.insert(self.playerBenefitCallbacks, AddCallback(ChatEventHandler, "PlayerBenefit", function(args)
            local value = data["PlayerBenefit"][args.SkillName]
            if value then
                valuebar:AddAlert(value["Icon"], value.Countdown, value.Text)
            end
        end))
    end
    
    self.currentHeight = self.currentHeight + data["Height"] + 2 * data["BorderSize"]

    table.insert(self.tools, valuebar)
end

function PrimeUI:Alert(data, parent)
    local x = data.X
    local y = data.Y
    if not(self.groupMode) then
        x = self:ContentAlignement(data["Alignment"], (data["Width"] + 2 * data["BorderSize"]) * data["Scale"])
        y = self.currentHeight
    end

    local valuebar = AlertPrimeUI(parent, x, y, data.Width, data.Height, data.BorderSize, data.Scale, self:ColourTupleToTurbineColour(data.BorderColour), self:ColourTupleToTurbineColour(data.BackgroundColour), self:ColourTupleToTurbineColour(data.Font_Colour), self:FontToFont(data.Font), Turbine.UI.ContentAlignment[data.TextAlignment], data.Text, self:FontToFont(data.Duration_Font), data.Show_Duration, Turbine.UI.ContentAlignment[data.DurationTextAlignment], data.Decimal_Threshold, self:ColourTupleToTurbineColour(data.Outline_Colour))

    if self.groupMode then
        valuebar.moveBox = MoveBox(valuebar, data["Width"], data["Height"], "Alert")
        valuebar.savePosition = function()
            if valuebar.moveBox.MoveBox:IsVisible() then
                valuebar:SetVisible(true)
            else
                valuebar:SetVisible(false)
            end
            data["X"], data["Y"] = valuebar:GetPosition()
        end
    end
    if tablelength(data["EffectAdded"]) > 0 then
        table.insert(self.effectAddedCallbacks, AddCallback(PlayerEffectsHandler, "EffectAdded", 
        function(effect)
            local value = data["EffectAdded"][effect:GetName()]
            if value then
                local time = value.Countdown
                if time == true then
                    time = effect:GetDuration() - (Turbine.Engine.GetGameTime() - effect:GetStartTime())
                end

                local icon
                if value["Show_Icon"] then
                    icon = value["Icon"]
                    if value["Icon"] == true then
                        icon = effect:GetIcon()
                    end
                end

                if value["Match_Icon"] then
                    if tostring(icon) ~= tostring(effect:GetIcon()) then return end
                end

                valuebar:SetAlert(time, value.Text, icon)
            end
        end))
    end

    if tablelength(data["EffectAdded_Target"]) > 0 then
        table.insert(self.effectTargetAddedCallbacks, TargetHandlerObject:AddEffectCallback( 
        function(effect)
            local value = data["EffectAdded_Target"][effect:GetName()]
            if value then
                local time = value.Countdown
                if time == true then
                    time = effect:GetDuration() - (Turbine.Engine.GetGameTime() - effect:GetStartTime())
                end

                
                local icon
                if value["Show_Icon"] then
                    icon = value["Icon"]
                    if value["Icon"] == true then
                        icon = effect:GetIcon()
                    end
                end

                if value["Match_Icon"] then
                    if tostring(icon) ~= tostring(effect:GetIcon()) then return end
                end

                valuebar:SetAlert(time, value.Text, icon)
            end
        end))
    end

    if tablelength(data["EffectAdded_Group"]) > 0 then
        table.insert(self.effectAddedGroupCallbacks, AddCallback(GroupEffectsHandler, "EffectAdded", 
        function(args)
            local value = data["EffectAdded_Group"][args.effect:GetName()]
            if value then
                local time = value.Countdown
                if time == true then
                    time = args.effect:GetDuration() - (Turbine.Engine.GetGameTime() - args.effect:GetStartTime())
                end

                local icon
                if value["Show_Icon"] then
                    icon = value["Icon"]
                    if value["Icon"] == true then
                        icon = args.effect:GetIcon()
                    end
                end

                if value["Match_Icon"] then
                    if tostring(icon) ~= tostring(args.effect:GetIcon()) then return end
                end

                valuebar:SetAlert(time, string.gsub(value.Text, "<name>", args.name ), icon)
            end
        end))
    end

    if tablelength(data["EffectRemoved"]) > 0 then
        table.insert(self.effectRemovedCallbacks, AddCallback(PlayerEffectsHandler, "EffectRemoved", 
        function(effect)
            local value = data["EffectRemoved"][effect:GetName()]
            if value then
                local time = value.Countdown
                if time == true then
                    time = effect:GetDuration() - (Turbine.Engine.GetGameTime() - effect:GetStartTime())
                end

                local icon
                if value["Show_Icon"] then
                    icon = value["Icon"]
                    if value["Icon"] == true then
                        icon = effect:GetIcon()
                    end
                end

                if value["Match_Icon"] then
                    if tostring(icon) ~= tostring(effect:GetIcon()) then return end
                end

                valuebar:SetAlert(time, value.Text, icon)
            end
        end))
    end

    if tablelength(data["Skills"]) > 0 then
        self.skills = true
        for key, value in pairs(data["Skills"]) do
            if SkillEventHandler:SkillExists(key) then
                -- Add the callback for this skill slot
                if not(self.skillCallbacks[key]) then self.skillCallbacks[key] = {} end
                table.insert(self.skillCallbacks[key], AddCallback(SkillEventHandler, key, function(skill) 
                    local time = value.Countdown
                    if time == true then
                        time = skill:GetResetTime() - Turbine.Engine.GetGameTime()
                    end
                    valuebar:SetAlert(time, value.Text, value["Icon"])
                end))
            end
        end
    end

    if tablelength(data["EnemyHits"]) > 0 then
        table.insert(self.enemyChatCallback, AddCallback(ChatEventHandler, "EnemyCombat", function(args)
            local value = data["EnemyHits"][args.SkillName]
            if value then
                valuebar:SetAlert(value.Countdown, value.Text, value["Icon"])
            end
        end))

    end

    if tablelength(data["PlayerHits"]) > 0 then
        table.insert(self.playerChatCallback, AddCallback(ChatEventHandler, "PlayerCombat", function(args)
            local value = data["PlayerHits"][args.SkillName]
            if value then
                valuebar:SetAlert(value.Countdown, value.Text, value["Icon"])
            end
        end))
    end

    if tablelength(data["StandardChat"]) > 0 then
        table.insert(self.standardChatCallback, AddCallback(ChatEventHandler, "Standard", function(message)
            local value = data["StandardChat"][message]
            if value then
                valuebar:SetAlert(value.Countdown, value.Text, value["Icon"])
            end
        end))
    end

    if tablelength(data["PartyChat"]) > 0 then
        table.insert(self.partyChatCallback, AddCallback(ChatEventHandler, "Party", function(message)
            local value = data["PartyChat"][message]
            if value then
                valuebar:SetAlert(value.Countdown, value.Text, value["Icon"])
            end
        end))
    end

    if tablelength(data["SayChat"]) > 0 then
        table.insert(self.sayChatCallback, AddCallback(ChatEventHandler, "SayLine", function(message)
            local value = data["SayChat"][message]
            if value then
                valuebar:SetAlert(value.Countdown, value.Text, value["Icon"])
            end
        end))
    end

    if tablelength(data["PlayerHeal"]) > 0 then
        table.insert(self.playerHealCallbacks, AddCallback(ChatEventHandler, "PlayerHeal", function(args)
            local value = data["PlayerHeal"][args.SkillName]
            if value then
                valuebar:SetAlert(value.Countdown, value.Text, value["Icon"])
            end
        end))
    end

    if tablelength(data["PlayerBenefit"]) > 0 then
        table.insert(self.playerBenefitCallbacks, AddCallback(ChatEventHandler, "PlayerBenefit", function(args)
            local value = data["PlayerBenefit"][args.SkillName]
            if value then
                valuebar:SetAlert(value.Countdown, value.Text, value["Icon"])
            end
        end))
    end
    
    self.currentHeight = self.currentHeight + data["Height"] + 2 * data["BorderSize"]

    table.insert(self.tools, valuebar)
end

function PrimeUI:Text(data, parent)
    local valuebar = Turbine.UI.Control()
    valuebar:SetParent(parent)
    if self.groupMode then
        valuebar:SetPosition(data["X"], data["Y"])
        valuebar.moveBox = MoveBox(valuebar, (data["Width"] + 2 * data["BorderSize"]) * data["Scale"], (data["Height"] + 2 * data["BorderSize"]) * data["Scale"], "Text")
        valuebar.savePosition = function()
            data["X"], data["Y"] = valuebar:GetPosition()
        end
    else
        valuebar:SetPosition(self:ContentAlignement(data["Alignment"], (data["Width"] + 2 * data["BorderSize"]) * data["Scale"]), self.currentHeight)
    end
    valuebar:SetSize((data["Width"] + 2 * data["BorderSize"]) * data["Scale"], (data["Height"] + 2 * data["BorderSize"]) * data["Scale"])
    valuebar:SetMouseVisible(false)
    valuebar:SetBackColor(self:ColourTupleToTurbineColour(data["BorderColour"]))

    valuebar.innerbar = Turbine.UI.Label()
    valuebar.innerbar:SetParent(valuebar)
    valuebar.innerbar:SetPosition(data["BorderSize"] * data["Scale"], data["BorderSize"] * data["Scale"])
    valuebar.innerbar:SetSize(data["Width"], data["Height"])
    valuebar.innerbar:SetMouseVisible(false)
    valuebar.innerbar:SetBlendMode(Turbine.UI.BlendMode.AlphaBlend)
    valuebar.innerbar:SetBackColor(self:ColourTupleToTurbineColour(data["BackgroundColour"]))
    valuebar.innerbar:SetForeColor(self:ColourTupleToTurbineColour(data["Font_Colour"]))
    valuebar.innerbar:SetFont(self:FontToFont(data["Font"]))
    valuebar.innerbar:SetText(data["Text"])

    valuebar.innerbar:SetStretchMode(1)
    valuebar.innerbar:SetSize(data["Width"] * data["Scale"], data["Height"] * data["Scale"])

    valuebar.baseText = data["Text"]
    valuebar.textTable = {}
    valuebar.instanceState = Settings.InstanceState

    valuebar.statusTable = {}
    for key, value in pairs(data["Effects"]) do
        if value == false then
            valuebar.statusTable[key] = 1
        else
            valuebar.statusTable[key] = 0
        end
    end

    valuebar.togglefunction = function()
        local tableValue = 0
        for key, value in pairs(valuebar.statusTable) do
            if value > 0 then
                tableValue = tableValue + 1
            end
        end
        local visibilityValue = tableValue >= valuebar.minCount
        valuebar:SetVisible(visibilityValue)
        if valuebar.innerbar.highlight then valuebar.innerbar.highlight:Toggle(visibilityValue) end
    end

    valuebar:SetVisible(false)
    valuebar.count = 0
    if data["Conjuction"] then
        valuebar.minCount = tablelength(data["Effects"]) + tablelength(data["HealthThreshold"]) + (data["Instance_Only"] and 1 or 0)
    else
        valuebar.minCount = 1
    end

    if data["Instance_Only"] and valuebar.instanceState then
        valuebar.count = valuebar.count + 1
    end

    if data["AlwaysVisible"] then
        valuebar:SetVisible(true)
    else
        for i = 1, self.player:GetEffects():GetCount(), 1 do
            local effect = self.player:GetEffects():Get(i)
            if data["Effects"][effect:GetName()] ~= nil then
                if data["Effects"][effect:GetName()] == true then
                    valuebar.statusTable[effect:GetName()] = valuebar.statusTable[effect:GetName()] + 1
                elseif data["Effects"][effect:GetName()] == false then
                    valuebar.statusTable[effect:GetName()] = valuebar.statusTable[effect:GetName()] - 1
                end
            end
        end


        if tablelength(data["Effects"]) > 0 then
            table.insert(self.effectAddedCallbacks, AddCallback(PlayerEffectsHandler, "EffectAdded", 
            function(effect)
                local name = effect:GetName()
                if data["Effects"][name] ~= nil then
                    local value = data["Effects"][effect:GetName()]
                    if value == true then
                        valuebar.statusTable[name] = valuebar.statusTable[name] + 1
                    else
                        valuebar.statusTable[name] = valuebar.statusTable[name] - 1
                    end
                    valuebar.togglefunction()
                end
            end))

            table.insert(self.effectRemovedCallbacks, AddCallback(PlayerEffectsHandler, "EffectRemoved", 
            function(effect)
                local name = effect:GetName()
                if data["Effects"][name] ~= nil then
                    local value = data["Effects"][effect:GetName()]
                    if value == true then
                        valuebar.statusTable[name] = valuebar.statusTable[name] - 1
                    else
                        valuebar.statusTable[name] = valuebar.statusTable[name] + 1
                    end
                    valuebar.togglefunction()
                end
            end))
        end

        if tablelength(data["HealthThreshold"]) > 0 then
            local health = self.player:GetMorale() / self.player:GetMaxMorale() * 100
            valuebar.healthThresholdPercentage = {}
            for key, value in pairs(data["HealthThreshold"]) do
                local checkValue = (self.compareFunctions[value["Operator"]](health, key) and 1) or 0
                valuebar.statusTable[tostring(key) .. tostring(value["Operator"])] = checkValue
                valuebar.healthThresholdPercentage[tostring(key) .. tostring(value["Operator"])] = checkValue
            end
            table.insert(self.playerHealthCallbacks, AddCallback(self.player, "MoraleChanged", function()
                local percentage = self.player:GetMorale() / self.player:GetMaxMorale() * 100
                local time = Turbine.Engine.GetGameTime()
                for key, value in pairs(data["HealthThreshold"]) do
                    local id = tostring(key) .. tostring(value["Operator"])
                    if self.compareFunctions[value["Operator"]](percentage, key) then
                        if valuebar.healthThresholdPercentage[id] == 0 then
                            if valuebar.healthThresholdPercentage[id] == 0 then
                                valuebar.healthThresholdPercentage[id] = 1
                                valuebar.statusTable[tostring(key) .. tostring(value["Operator"])] = valuebar.statusTable[tostring(key) .. tostring(value["Operator"])] + 1
                            end
                        end
                    else
                        if valuebar.healthThresholdPercentage[id] == 1 then
                            valuebar.healthThresholdPercentage[id] = 0
                            valuebar.statusTable[tostring(key) .. tostring(value["Operator"])] = valuebar.statusTable[tostring(key) .. tostring(value["Operator"])] - 1
                        end
                    end
                end
                valuebar.togglefunction()
            end))
        end
        valuebar.togglefunction()
    end

    if true then
        --[[
            Instances:
            State - In an instance or not
            Instance_Name - Name of the instance or all
            Instance_Tier - Tier of instance or \"\"

            Boss fights:
            Boss_Name - Name of the boss fight
            Boss_Tier - Tier of the boss fight
            Boss_Completion_Time - Fight duration (wipe or completion)
            Boss_Wipes - Amount of wipes in a fight

            Player:
            Player_Name - Name of the player
            Player_Class - Class of the player

            Stats:
            Death_Count_Character - Amount of deaths on character
            Death_Count_Account - Amount of deaths on account
            Death_Count_Misadventure_Character - Amount of Misadventure deaths on character
            Death_Count_Misadventure_Account - Amount of Misadventure deaths on account

        ]]
        local startData = InstanceController:GetInitialData()
        local charData, accData = StatEventHandling:GetData()

        startData["Player_Name"] = self.player:GetName()
        startData["Player_Class"] = self.classObjectsToName[self.player:GetClass()]
        startData["Death_Count_Character"] = tostring(charData["Death"])
        startData["Death_Count_Account"] = tostring(accData["Death"])
        startData["Death_Count_Misadventure_Character"] = tostring(charData["MisadventureDeaths"])
        startData["Death_Count_Misadventure_Account"] = tostring(accData["MisadventureDeaths"])

        
        for key, value in pairs(startData) do
            if key == "Instance_Tier" and value == 0 then startData[key] = nil end
            if key == "Boss_Tier" and value == 0 then startData[key] = nil end
            if key == "Instance_Name" and value == "All" then startData[key] = nil end
        end


        self:SetText(valuebar.innerbar, valuebar.textTable, startData, valuebar.baseText)  

        table.insert(self.deathChatCallbacks, AddCallback(ChatEventHandler, "Death", function(args)
            local charData, accData = StatEventHandling:GetData()
            self:SetText(valuebar.innerbar, valuebar.textTable, 
            {["Death_Count_Character"] = tostring(charData["Death"]), 
            ["Death_Count_Account"] = tostring(accData["Death"]),
            ["Death_Count_Misadventure_Character"] = tostring(charData["MisadventureDeaths"]),
            ["Death_Count_Misadventure_Account"] = tostring(accData["MisadventureDeaths"])}, valuebar.baseText)   
        end))

        -- Instance callback
        table.insert(self.instanceEventCallbacks, AddCallback(InstanceController, "InstanceEvent", function(args)
            if args["Instance_Tier"] == 0 then args["Instance_Tier"] = nil end
            if args["Instance_Name"] == "All" then args["Instance_Name"] = nil end

            self:SetText(valuebar.innerbar, valuebar.textTable, args, valuebar.baseText)   
            
            if data["Instance_Only"] then
                if args["State"] and not(valuebar.instanceState) then
                    valuebar.count = valuebar.count + 1
                    valuebar.instanceState = true
                elseif valuebar.instanceState then
                    valuebar.count = valuebar.count - 1
                    valuebar.instanceState = false
                end
                valuebar:SetVisible(valuebar.count == valuebar.minCount)
            end
        end))

        -- Fight callback
        table.insert(self.fightEventStartCallbacks, AddCallback(InstanceController, "FightEventStart", function(args)
            self:SetText(valuebar.innerbar, valuebar.textTable, args, valuebar.baseText)  
        end))

        table.insert(self.fightEventEndCallbacks, AddCallback(InstanceController, "FightEventEnd", function(args)
            self:SetText(valuebar.innerbar, valuebar.textTable, args, valuebar.baseText)  
        end))
    end

    
    self.currentHeight = self.currentHeight + (data["Height"] + 2 * data["BorderSize"]) * data["Scale"]

    table.insert(self.UI, valuebar.innerbar)
    table.insert(self.UI, valuebar)
end

function PrimeUI:SetText(label, table, args, base)
    for key, value in pairs(args) do
        table["<" .. tostring(key) .. ">"] = tostring(value)
    end

    local newText = base
    for key, value in pairs(table) do
        newText = newText:gsub(key, value)
    end


    
    newText = newText:gsub("<.->%[.-%]", "")
    
    newText = newText:gsub("%[.-%]<.->", "")
    
    newText = newText:gsub("%[", "")
    
    newText = newText:gsub("%]", "")
    
    newText = newText:gsub("<.->", "")

    

    

    label:SetText(newText)
end

function PrimeUI:SetupSkillBar(data, parent)
    -- Determine bar height
    local barHeight = 0
    self.skills = true

    -- Bar -> Bar name
    -- Bar data
    -- Skills
    -- Scale

    --args (parent, x, y, width, height, labelSize, sortMethod, hide)
    local skillbar = DynamicSkillBar(parent, 0, self.currentHeight, self.width, 34 * data["Scale"], 34 * data["Scale"], 4, false, false, data["Offset"])
    table.insert(self.tools, skillbar)

    -- Skill data
    -- Name
    -- ID
    -- X prio
    -- Y prio
    -- Special triggers?
    local count = 0
    local height = 1

    -- [[
    local highlightTable = {}

    if data["Highlight_On_Effect"] then
        for key, value in pairs(data["Highlight_On_Effect"]) do
            highlightTable[value.Skill] = true
        end
    end
    --]]

    for key, value in pairs(data["Skills"]) do

        if SkillEventHandler:SkillExists(key) then

            local highlight = highlightTable[key]

            --args (parent, x, y, base, normal, skill, hideOnCreation, pet, highlight)
            local skillSlot = ClassSkillSlotPrimeUI(skillbar, 0, 0, 34, 34, value["Code"], self:ColourTupleToTurbineColour(data["BorderColour"]), data["BorderSize"], nil, 
            value["visibitilityThreshold"], value["hideWhileOff"], value["hideWhileOn"], value["opacityOn"], value["opacityOff"], self:FontToFont(data["Font"]), data["Scale"], highlight, self:ColourTupleToTurbineColour(data["HighlightColour"]), data["HighlightSize"]) --data["Scale"])

            count = math.max(count, value["XPrio"])
            
            -- Add the callback for this skill slot
            if not(self.skillCallbacks[key]) then self.skillCallbacks[key] = {} end
            table.insert(self.skillCallbacks[key], AddCallback(SkillEventHandler, key, function(skill) 
                local time = skill:GetResetTime() - Turbine.Engine.GetGameTime()
                if time + 1 > skill:GetCooldown() then
                    
                    if isnumber(value["Time"]) then
                        time = value["Time"]
                    end
                end
                skillbar:SetLabelTimer(key, math.max(0.1, time), skill:GetCooldown())
            end))

            -- args (skill, label, priority, skillObject, y)
            skillbar:AddLabel(key, skillSlot, value["XPrio"], nil, value["YPrio"])

            -- Update max bar height
            barHeight = math.max(barHeight, value["YPrio"])
        end
    end

    if tablelength(data["Quickslots"]) > 0 then
        for key, value in pairs(data["Quickslots"]) do
            local skillSlot = ClassSkillSlotPrimeUI(skillbar, 0, 0, 34, 34, {["Data"] = value["Data"], ["Type"] = value["Type"], ["Item"] = value["Item"]}, self:ColourTupleToTurbineColour(data["BorderColour"]), data["BorderSize"], nil, 
            value["visibitilityThreshold"], value["hideWhileOff"], value["hideWhileOn"], value["opacityOn"], value["opacityOff"], self:FontToFont(data["Font"]), data["Scale"], nil, self:ColourTupleToTurbineColour(data["HighlightColour"]), data["HighlightSize"]) --data["Scale"])

            count = math.max(count, value["XPrio"])

            -- args (skill, label, priority, skillObject, y)
            skillbar:AddLabel(key, skillSlot, value["XPrio"], nil, value["YPrio"])

            -- Update max bar height
            barHeight = math.max(barHeight, value["YPrio"])
        end
    end

    if tablelength(data["Highlight_On_Effect"]) > 0 then
        skillbar.SkillID = {}
        skillbar.SkillTable = {}
        skillbar.EffectIDs = {}
        for key, value in pairs(data["Highlight_On_Effect"]) do
            if not skillbar.SkillTable[value["Effect"]] then skillbar.SkillTable[value["Effect"]] = {} end
            table.insert(skillbar.SkillTable[value["Effect"]], key)
        end

        local effectsFound = {}

        for i = 1, self.player:GetEffects():GetCount() do
            local effect = self.player:GetEffects():Get(i)
            if skillbar.SkillTable[effect:GetName()] then
                effectsFound[effect:GetName()] = true
                skillbar.EffectIDs[effect:GetName()] = effect:GetID()
                for _, index in pairs(skillbar.SkillTable[effect:GetName()]) do
                    local value = data["Highlight_On_Effect"][index]
                    local ID = effect:GetID()
    
                    skillbar.SkillID[ID] = true
                    skillbar:ToggleHighlight(value["Skill"], not(value["Invert_Highlight"]))
                end
            end
        end

        for key, _ in pairs(skillbar.SkillTable) do
            if not(effectsFound[key]) then
                for _, index in pairs(skillbar.SkillTable[key]) do
                    local value = data["Highlight_On_Effect"][index]

                    if value["Invert_Highlight"] then
                        skillbar:ToggleHighlight(value["Skill"], value["Invert_Highlight"])
                    end
                end
            end
        end

        table.insert(self.effectAddedCallbacks, AddCallback(PlayerEffectsHandler, "EffectAdded", 
        function(effect)
            local name = effect:GetName()
            if skillbar.SkillTable[name] then
                skillbar.EffectIDs[effect:GetName()] = effect:GetID()
                for _, index in pairs(skillbar.SkillTable[name]) do
                    local value = data["Highlight_On_Effect"][index]
                    local ID = effect:GetID()
    
                    skillbar.SkillID[ID] = true
                    skillbar:ToggleHighlight(value["Skill"], not(value["Invert_Highlight"]))
                end
            end
        end))

        table.insert(self.effectRemovedCallbacks, AddCallback(PlayerEffectsHandler, "EffectRemoved", 
        function(effect)
            local ID = effect:GetID()
            if skillbar.SkillID[ID] and skillbar.EffectIDs[effect:GetName()] == effect:GetID() then
                for _, index in pairs(skillbar.SkillTable[effect:GetName()]) do
                    local value = data["Highlight_On_Effect"][index]
                    skillbar:ToggleHighlight(value["Skill"], value["Invert_Highlight"])
                end
            end

        end))
    end

    skillbar:SetStaticPositions()

    data["Width"] = (35 + (data["Offset"] or 1)) * count

    if self.groupMode then
        skillbar.moveBox = MoveBox(skillbar, (35 + (data["Offset"] or 1)) * count + 50, barHeight * 34 * data["Scale"] + 50, "Skill bar")
        skillbar.savePosition = function()
            data["X"], data["Y"] = skillbar:GetPosition()
        end
    end

    -- Update height
    self.currentHeight = self.currentHeight + barHeight * 34 * data["Scale"] + barHeight * (data["BorderSize"] or 0)
end

function PrimeUI:ContentAlignement(alignment, width)
    if alignment == "Middle" then
        return (self.width - width) / 2
    elseif alignment == "Right" then
        return self.width - width
    else
        return 0 
    end
end

function PrimeUI:Countdown(data, parent)
    local valuebar = Turbine.UI.Control()
    valuebar:SetParent(parent)
    if self.groupMode then
        valuebar:SetPosition(data["X"], data["Y"])
        valuebar.moveBox = MoveBox(valuebar, ((data["Width"] or 100) * (data["Scale"] or 1)) or 100, ((data["Height"] or 50) * (data["Scale"] or 1)) or 50, "Countdown")
        valuebar.savePosition = function()
            data["X"], data["Y"] = valuebar:GetPosition()
        end
    else
        valuebar:SetPosition(self:ContentAlignement((data["Alignment"] or "Middle"), (data["Width"] or 100) * (data["Scale"] or 1)), self.currentHeight)
    end
    valuebar:SetSize((data["Width"] or 100) * (data["Scale"] or 1), (data["Height"] or 50) * (data["Scale"] or 1))
    valuebar.countdownWindow = CountdownWindowPrimeUI(valuebar, 0, 0, data["Width"], data["Height"], self:FontToFont(data["Font"]), data["Scale"], self:ColourTupleToTurbineColour(data["FontColour"]), self:ColourTupleToTurbineColour(data["ThresholdColour"]), data["Threshold"])

    if tablelength(data["EffectAdded"]) > 0 then
        table.insert(self.effectAddedCallbacks, AddCallback(PlayerEffectsHandler, "EffectAdded", 
        function(effect)
            local name = effect:GetName()
            local time = data["EffectAdded"][name]
            if time then
                if time == true then
                    time = effect:GetDuration() - (Turbine.Engine.GetGameTime() - effect:GetStartTime())
                end
                valuebar.countdownWindow:Countdown(time, data["Visibility_Threshold"])
            end
        end))
    end

    if tablelength(data["EffectAdded_Target"]) > 0 then
        table.insert(self.effectTargetAddedCallbacks, TargetHandlerObject:AddEffectCallback(
        function(effect)
            local name = effect:GetName()
            local time = data["EffectAdded_Target"][name]
            if time then
                if time == true then
                    time = effect:GetDuration() - (Turbine.Engine.GetGameTime() - effect:GetStartTime())
                end
                valuebar.countdownWindow:Countdown(time, data["Visibility_Threshold"])
            end
        end))
    end

    if tablelength(data["EffectAdded_Group"]) > 0 then
        table.insert(self.effectAddedGroupCallbacks, AddCallback(GroupEffectsHandler, "EffectAdded", 
        function(args)
            local time = data["EffectAdded_Group"][args.effect:GetName()]
            if time then
                if time == true then
                    time = args.effect:GetDuration() - (Turbine.Engine.GetGameTime() - args.effect:GetStartTime())
                end
                valuebar.countdownWindow:Countdown(time, data["Visibility_Threshold"])
            end
        end))
    end

    if tablelength(data["EffectRemoved"]) > 0 then
        table.insert(self.effectRemovedCallbacks, AddCallback(PlayerEffectsHandler, "EffectRemoved", 
        function(effect)
            local name = effect:GetName()
            local time = data["EffectRemoved"][name]
            if time then
                if time == true then
                    time = effect:GetDuration() - (Turbine.Engine.GetGameTime() - effect:GetStartTime())
                end
                valuebar.countdownWindow:Countdown(time, data["Visibility_Threshold"])
            end
        end))
    end

    if tablelength(data["EnemyHits"]) > 0 then
        table.insert(self.enemyChatCallback, AddCallback(ChatEventHandler, "EnemyCombat", function(args)
            local value = data["EnemyHits"][args.SkillName]
            if value then
                valuebar.countdownWindow:Countdown(value, data["Visibility_Threshold"])
            end
        end))

    end

    if tablelength(data["PlayerHits"]) > 0 then
        table.insert(self.playerChatCallback, AddCallback(ChatEventHandler, "PlayerCombat", function(args)
            local value = data["PlayerHits"][args.SkillName]
            if value then
                valuebar.countdownWindow:Countdown(value, data["Visibility_Threshold"])
            end
        end))
    end

    if tablelength(data["StandardChat"]) > 0 then
        table.insert(self.standardChatCallback, AddCallback(ChatEventHandler, "Standard", function(message)
            if data["StandardChat"][message] then
                valuebar.countdownWindow:Countdown(data["StandardChat"][message], data["Visibility_Threshold"])
            end
        end))
    end

    if tablelength(data["PartyChat"]) > 0 then
        table.insert(self.partyChatCallback, AddCallback(ChatEventHandler, "Party", function(message)
            if data["PartyChat"][message] then
                valuebar.countdownWindow:Countdown(data["PartyChat"][message], data["Visibility_Threshold"])
            end
        end))
    end

    if tablelength(data["SayChat"]) > 0 then
        table.insert(self.sayChatCallback, AddCallback(ChatEventHandler, "SayLine", function(message)
            if data["SayChat"][message]  then
                valuebar.countdownWindow:Countdown(data["SayChat"][message], data["Visibility_Threshold"])
            end
        end))
    end

    if tablelength(data["PlayerBenefit"]) > 0 then
        table.insert(self.playerBenefitCallbacks, AddCallback(ChatEventHandler, "PlayerBenefit", function(args)
            if data["PlayerBenefit"][args.SkillName]  then
                valuebar.countdownWindow:Countdown(data["PlayerBenefit"][args.SkillName], data["Visibility_Threshold"])
            end
        end))
    end

    if tablelength(data["PlayerHeal"]) > 0 then
        table.insert(self.playerHealCallbacks, AddCallback(ChatEventHandler, "PlayerHeal", function(args)
            if data["PlayerHeal"][args.SkillName]  then
                valuebar.countdownWindow:Countdown(data["PlayerHeal"][args.SkillName], data["Visibility_Threshold"])
            end
        end))
    end


    if tablelength(data["Skills"]) > 0 then
        self.skills = true
        for key, value in pairs(data["Skills"]) do
            if SkillEventHandler:SkillExists(key) then
                -- Add the callback for this skill slot
                if not(self.skillCallbacks[key]) then self.skillCallbacks[key] = {} end
                table.insert(self.skillCallbacks[key], AddCallback(SkillEventHandler, key, function(skill) 
                    local time = data["Skills"][key]
                    if time then
                        if time == true then
                            time = skill:GetResetTime() - Turbine.Engine.GetGameTime()
                        end
                        valuebar.countdownWindow:Countdown(time, data["Visibility_Threshold"])
                    end
                end))
            end
        end
    end

    if tablelength(data["Target_HealthThreshold"]) > 0 then
        valuebar.targetHealthThresholdCooldowns = {}
        valuebar.targetHealthThresholdPercentage = {}
        valuebar.targetMaxHealth = 0
        local health = self.player:GetMorale() / self.player:GetMaxMorale() * 100
        for key, value in pairs(data["Target_HealthThreshold"]) do
            valuebar.targetHealthThresholdCooldowns[key .. tostring(value["Number"]) .. tostring(value["Operator"])] = Turbine.Engine.GetGameTime() - 1000
            valuebar.targetHealthThresholdPercentage[key .. tostring(value["Number"]) .. tostring(value["Operator"])] = true
        end

        table.insert(self.targetInfoCallbacks, TargetHandlerObject:AddTargetInfoCallback(function(value) 
            if value then
                valuebar.targetMaxHealth = value["MaxMorale"]
                valuebar.TargetName = value["Name"]
            else
                valuebar.targetMaxHealth = 0
                valuebar.TargetName = ""
            end
        end))

        table.insert(self.healthTargetCallbacks, TargetHandlerObject:AddHealthCallback(function(value)
            if valuebar.targetMaxHealth > 0 then
                local percentage = value / valuebar.targetMaxHealth * 100
                local time = Turbine.Engine.GetGameTime()
                for key, value in pairs(data["Target_HealthThreshold"]) do
                    if key == valuebar.TargetName then
                        if self.compareFunctions[value["Operator"]](percentage, value["Number"]) then
                            local id = key .. tostring(value["Number"]) .. tostring(value["Operator"])
                            if valuebar.targetHealthThresholdCooldowns[id] < time and valuebar.targetHealthThresholdPercentage[id] then
                                valuebar.targetHealthThresholdPercentage[id] = false
                                valuebar.countdownWindow:Countdown(value["Countdown"], data["Visibility_Threshold"])
                                valuebar.targetHealthThresholdCooldowns[id] = time + 6
                            end
                        else
                            valuebar.targetHealthThresholdPercentage[key .. tostring(value["Number"]) .. tostring(value["Operator"])] = true
                        end
                    end
                end
            end
        end))
        table.insert(self.combatCallbacks, AddCallback(self.player, "InCombatChanged", function()
            if not(self.player:IsInCombat()) then
                for key, value in pairs(data["Target_HealthThreshold"]) do
                    valuebar.targetHealthThresholdCooldowns[key .. tostring(value["Number"]) .. tostring(value["Operator"])] = Turbine.Engine.GetGameTime() - 1000
                    valuebar.targetHealthThresholdPercentage[key .. tostring(value["Number"]) .. tostring(value["Operator"])] = true
                end 
            end
        end))
    end

    if not(data["Height"]) then
        data["Width"] = 100
        data["Height"] = 50
    end

    self.currentHeight = self.currentHeight + (data["Height"] or 125) * (data["Scale"] or 1)

    table.insert(self.tools, valuebar.countdownWindow)
    table.insert(self.UI, valuebar)
    
end

function PrimeUI:Sound(data, parent)
    local valuebar = Turbine.UI.Control()
    valuebar:SetParent(parent)
    valuebar:SetVisible(false)

    if tablelength(data["EffectAdded"]) > 0 then
        table.insert(self.effectAddedCallbacks, AddCallback(PlayerEffectsHandler, "EffectAdded", 
        function(effect)
            local name = effect:GetName()
            if data["EffectAdded"][name] ~= nil then
                Sounds:PlaySoundEffect(SoundEffects[data["EffectAdded"][name]])
            end
        end))
    end

    if tablelength(data["EffectAdded_Target"]) > 0 then
        table.insert(self.effectTargetAddedCallbacks, TargetHandlerObject:AddEffectCallback(
        function(effect)
            local name = effect:GetName()
            if data["EffectAdded_Target"][name] ~= nil then
                Sounds:PlaySoundEffect(SoundEffects[data["EffectAdded_Target"][name]])
            end
        end))
    end

    if tablelength(data["EffectAdded_Group"]) > 0 then
        table.insert(self.effectAddedGroupCallbacks, AddCallback(GroupEffectsHandler, "EffectAdded", 
        function(args)
            local name = args.effect:GetName()
            if data["EffectAdded_Group"][name] ~= nil then
                Sounds:PlaySoundEffect(SoundEffects[data["EffectAdded_Group"][name]])
            end
        end))
    end

    if tablelength(data["EffectRemoved"]) > 0 then
        table.insert(self.effectRemovedCallbacks, AddCallback(PlayerEffectsHandler, "EffectRemoved", 
        function(effect)
            local name = effect:GetName()
            if data["EffectRemoved"][name] ~= nil then
                Sounds:PlaySoundEffect(SoundEffects[data["EffectRemoved"][name]])
            end
        end))
    end

    if tablelength(data["EnemyHits"]) > 0 then
        table.insert(self.enemyChatCallback, AddCallback(ChatEventHandler, "EnemyCombat", function(args)
            local value = data["EnemyHits"][args.SkillName]
            if value then
                Sounds:PlaySoundEffect(SoundEffects[value])
            end
        end))

    end

    if tablelength(data["StandardChat"]) > 0 then
        table.insert(self.standardChatCallback, AddCallback(ChatEventHandler, "Standard", function(message)
            if data["StandardChat"][message] then
                Sounds:PlaySoundEffect(SoundEffects[data["StandardChat"][message]])
            end
        end))
    end

    if tablelength(data["PartyChat"]) > 0 then
        table.insert(self.partyChatCallback, AddCallback(ChatEventHandler, "Party", function(message)
            if data["PartyChat"][message] then
                Sounds:PlaySoundEffect(SoundEffects[data["PartyChat"][message]])
            end
        end))
    end

    if tablelength(data["SayChat"]) > 0 then
        table.insert(self.sayChatCallback, AddCallback(ChatEventHandler, "SayLine", function(message)
            if data["SayChat"][message]  then
                Sounds:PlaySoundEffect(SoundEffects[data["SayChat"][message]])
            end
        end))
    end

    if tablelength(data["PlayerHits"]) > 0 then
        table.insert(self.playerChatCallback, AddCallback(ChatEventHandler, "PlayerCombat", function(args)
            local value = data["PlayerHits"][args.SkillName]
            if value then
                Sounds:PlaySoundEffect(SoundEffects[value])
            end
        end))
    end

    if tablelength(data["PlayerBenefit"]) > 0 then
        table.insert(self.playerBenefitCallbacks, AddCallback(ChatEventHandler, "PlayerBenefit", function(args)
            local value = data["PlayerBenefit"][args.SkillName]
            if value then
                Sounds:PlaySoundEffect(SoundEffects[value])
            end
        end))
    end

    if tablelength(data["PlayerHeal"]) > 0 then
        table.insert(self.playerHealCallbacks, AddCallback(ChatEventHandler, "PlayerHeal", function(args)
            local value = data["PlayerHeal"][args.SkillName]
            if value then
                Sounds:PlaySoundEffect(SoundEffects[value])
            end
        end))
    end

    if tablelength(data["HealthThreshold"]) > 0 then
        local health = self.player:GetMorale() / self.player:GetMaxMorale() * 100
        for key, value in pairs(data["HealthThreshold"]) do
            self.healthThresholdCooldowns[tostring(key) .. tostring(value["Operator"])] = Turbine.Engine.GetGameTime() - 1000
            self.healthThresholdPercentage[tostring(key) .. tostring(value["Operator"])] = not(self.compareFunctions[value["Operator"]](health, key))
        end
        table.insert(self.playerHealthCallbacks, AddCallback(self.player, "MoraleChanged", function()
            local percentage = self.player:GetMorale() / self.player:GetMaxMorale() * 100
            local time = Turbine.Engine.GetGameTime()
            for key, value in pairs(data["HealthThreshold"]) do
                if self.compareFunctions[value["Operator"]](percentage, key) then
                    local id = tostring(key) .. tostring(value["Operator"])
                    if self.healthThresholdCooldowns[id] < time and self.healthThresholdPercentage[id] then
                        self.healthThresholdPercentage[id] = false
                        Sounds:PlaySoundEffect(SoundEffects[value["Sound"]])
                        self.healthThresholdCooldowns[id] = time + 6
                    end
                else
                    self.healthThresholdPercentage[tostring(key) .. tostring(value["Operator"])] = true
                end
                
            end
        end))
    end

    if tablelength(data["Target_HealthThreshold"]) > 0 then
        valuebar.targetHealthThresholdCooldowns = {}
        valuebar.targetHealthThresholdPercentage = {}
        valuebar.targetMaxHealth = 0
        local health = self.player:GetMorale() / self.player:GetMaxMorale() * 100
        for key, value in pairs(data["Target_HealthThreshold"]) do
            valuebar.targetHealthThresholdCooldowns[key .. tostring(value["Number"]) .. tostring(value["Operator"])] = Turbine.Engine.GetGameTime() - 1000
            valuebar.targetHealthThresholdPercentage[key .. tostring(value["Number"]) .. tostring(value["Operator"])] = true
        end

        table.insert(self.targetInfoCallbacks, TargetHandlerObject:AddTargetInfoCallback(function(value) 
            if value then
                valuebar.targetMaxHealth = value["MaxMorale"]
                valuebar.TargetName = value["Name"]
            else
                valuebar.targetMaxHealth = 0
                valuebar.TargetName = ""
            end
        end))

        table.insert(self.healthTargetCallbacks, TargetHandlerObject:AddHealthCallback(function(value)
            if valuebar.targetMaxHealth > 0 then
                local percentage = value / valuebar.targetMaxHealth * 100
                local time = Turbine.Engine.GetGameTime()
                for key, value in pairs(data["Target_HealthThreshold"]) do
                    if key == valuebar.TargetName then
                        if self.compareFunctions[value["Operator"]](percentage, value["Number"]) then
                            local id = key .. tostring(value["Number"]) .. tostring(value["Operator"])
                            if valuebar.targetHealthThresholdCooldowns[id] < time and valuebar.targetHealthThresholdPercentage[id] then
                                valuebar.targetHealthThresholdPercentage[id] = false
                                Sounds:PlaySoundEffect(SoundEffects[value["Sound"]])
                                valuebar.targetHealthThresholdCooldowns[id] = time + 6
                            end
                        else
                            valuebar.targetHealthThresholdPercentage[key .. tostring(value["Number"]) .. tostring(value["Operator"])] = true
                        end
                    end
                end
            end
        end))
        table.insert(self.combatCallbacks, AddCallback(self.player, "InCombatChanged", function()
            if not(self.player:IsInCombat()) then
                for key, value in pairs(data["Target_HealthThreshold"]) do
                    valuebar.targetHealthThresholdCooldowns[key .. tostring(value["Number"]) .. tostring(value["Operator"])] = Turbine.Engine.GetGameTime() - 1000
                    valuebar.targetHealthThresholdPercentage[key .. tostring(value["Number"]) .. tostring(value["Operator"])] = true
                end 
            end
        end))
    end

    if tablelength(data["Skills"]) > 0 then
        self.skills = true
        for key, value in pairs(data["Skills"]) do
            if SkillEventHandler:SkillExists(key) then
                -- Add the callback for this skill slot
                if not(self.skillCallbacks[key]) then self.skillCallbacks[key] = {} end
                table.insert(self.skillCallbacks[key], AddCallback(SkillEventHandler, key, function(skill) 
                    Sounds:PlaySoundEffect(SoundEffects[data["Skills"][key]])
                end))
            end
        end
    end

    table.insert(self.UI, valuebar)
end

function PrimeUI:NoteEvent(data, parent)
    local valuebar = Turbine.UI.Control()
    valuebar:SetParent(parent)
    valuebar:SetVisible(false)

    if tablelength(data["EffectAdded"]) > 0 then
        table.insert(self.effectAddedCallbacks, AddCallback(PlayerEffectsHandler, "EffectAdded", 
        function(effect)
            local name = effect:GetName()
            if data["EffectAdded"][name] ~= nil then
                NoteHandler:HanldeEvent(data["EffectAdded"][name])
            end
        end))
    end

    if tablelength(data["EffectAdded_Target"]) > 0 then
        table.insert(self.effectTargetAddedCallbacks, TargetHandlerObject:AddEffectCallback(
        function(effect)
            local name = effect:GetName()
            if data["EffectAdded_Target"][name] ~= nil then
                NoteHandler:HanldeEvent(data["EffectAdded_Target"][name])
            end
        end))
    end

    if tablelength(data["EffectAdded_Group"]) > 0 then
        table.insert(self.effectAddedGroupCallbacks, AddCallback(GroupEffectsHandler, "EffectAdded", 
        function(args)
            local name = args.effect:GetName()
            if data["EffectAdded_Group"][name] ~= nil then
                NoteHandler:HanldeEvent(data["EffectAdded_Group"][name])
            end
        end))
    end

    if tablelength(data["EffectRemoved"]) > 0 then
        table.insert(self.effectRemovedCallbacks, AddCallback(PlayerEffectsHandler, "EffectRemoved", 
        function(effect)
            local name = effect:GetName()
            if data["EffectRemoved"][name] ~= nil then
                NoteHandler:HanldeEvent(data["EffectRemoved"][name])
            end
        end))
    end

    if tablelength(data["EnemyHits"]) > 0 then
        table.insert(self.enemyChatCallback, AddCallback(ChatEventHandler, "EnemyCombat", function(args)
            local value = data["EnemyHits"][args.SkillName]
            if value then
                NoteHandler:HanldeEvent(value)
            end
        end))

    end

    if tablelength(data["StandardChat"]) > 0 then
        table.insert(self.standardChatCallback, AddCallback(ChatEventHandler, "Standard", function(message)
            if data["StandardChat"][message] then
                NoteHandler:HanldeEvent(data["StandardChat"][message])
            end
        end))
    end

    if tablelength(data["PartyChat"]) > 0 then
        table.insert(self.partyChatCallback, AddCallback(ChatEventHandler, "Party", function(message)
            if data["PartyChat"][message] then
                NoteHandler:HanldeEvent(data["PartyChat"][message])
            end
        end))
    end

    if tablelength(data["SayChat"]) > 0 then
        table.insert(self.sayChatCallback, AddCallback(ChatEventHandler, "SayLine", function(message)
            if data["SayChat"][message]  then
                NoteHandler:HanldeEvent(data["SayChat"][message])
            end
        end))
    end

    if tablelength(data["PlayerHits"]) > 0 then
        table.insert(self.playerChatCallback, AddCallback(ChatEventHandler, "PlayerCombat", function(args)
            local value = data["PlayerHits"][args.SkillName]
            if value then
                NoteHandler:HanldeEvent(value)
            end
        end))
    end

    if tablelength(data["PlayerBenefit"]) > 0 then
        table.insert(self.playerBenefitCallbacks, AddCallback(ChatEventHandler, "PlayerBenefit", function(args)
            local value = data["PlayerBenefit"][args.SkillName]
            if value then
                NoteHandler:HanldeEvent(value)
            end
        end))
    end

    if tablelength(data["PlayerHeal"]) > 0 then
        table.insert(self.playerHealCallbacks, AddCallback(ChatEventHandler, "PlayerHeal", function(args)
            local value = data["PlayerHeal"][args.SkillName]
            if value then
                NoteHandler:HanldeEvent(value)
            end
        end))
    end

    if tablelength(data["HealthThreshold"]) > 0 then
        local health = self.player:GetMorale() / self.player:GetMaxMorale() * 100
        for key, value in pairs(data["HealthThreshold"]) do
            self.healthThresholdCooldowns[tostring(key) .. tostring(value["Operator"])] = Turbine.Engine.GetGameTime() - 1000
            self.healthThresholdPercentage[tostring(key) .. tostring(value["Operator"])] = not(self.compareFunctions[value["Operator"]](health, key))
        end
        table.insert(self.playerHealthCallbacks, AddCallback(self.player, "MoraleChanged", function()
            local percentage = self.player:GetMorale() / self.player:GetMaxMorale() * 100
            local time = Turbine.Engine.GetGameTime()
            for key, value in pairs(data["HealthThreshold"]) do
                if self.compareFunctions[value["Operator"]](percentage, key) then
                    local id = tostring(key) .. tostring(value["Operator"])
                    if self.healthThresholdCooldowns[id] < time and self.healthThresholdPercentage[id] then
                        self.healthThresholdPercentage[id] = false
                        NoteHandler:HanldeEvent(value["Event"])
                        self.healthThresholdCooldowns[id] = time + 6
                    end
                else
                    self.healthThresholdPercentage[tostring(key) .. tostring(value["Operator"])] = true
                end
                
            end
        end))
    end

    if tablelength(data["Target_HealthThreshold"]) > 0 then
        valuebar.targetHealthThresholdCooldowns = {}
        valuebar.targetHealthThresholdPercentage = {}
        valuebar.targetMaxHealth = 0
        local health = self.player:GetMorale() / self.player:GetMaxMorale() * 100
        for key, value in pairs(data["Target_HealthThreshold"]) do
            valuebar.targetHealthThresholdCooldowns[key .. tostring(value["Number"]) .. tostring(value["Operator"])] = Turbine.Engine.GetGameTime() - 1000
            valuebar.targetHealthThresholdPercentage[key .. tostring(value["Number"]) .. tostring(value["Operator"])] = true
        end

        table.insert(self.targetInfoCallbacks, TargetHandlerObject:AddTargetInfoCallback(function(value) 
            if value then
                valuebar.targetMaxHealth = value["MaxMorale"]
                valuebar.TargetName = value["Name"]
            else
                valuebar.targetMaxHealth = 0
                valuebar.TargetName = ""
            end
        end))

        table.insert(self.healthTargetCallbacks, TargetHandlerObject:AddHealthCallback(function(value)
            if valuebar.targetMaxHealth > 0 then
                local percentage = value / valuebar.targetMaxHealth * 100
                local time = Turbine.Engine.GetGameTime()
                for key, value in pairs(data["Target_HealthThreshold"]) do
                    if key == valuebar.TargetName then
                        if self.compareFunctions[value["Operator"]](percentage, value["Number"]) then
                            local id = key .. tostring(value["Number"]) .. tostring(value["Operator"])
                            if valuebar.targetHealthThresholdCooldowns[id] < time and valuebar.targetHealthThresholdPercentage[id] then
                                valuebar.targetHealthThresholdPercentage[id] = false
                                NoteHandler:HanldeEvent(value["Event"])
                                valuebar.targetHealthThresholdCooldowns[id] = time + 6
                            end
                        else
                            valuebar.targetHealthThresholdPercentage[key .. tostring(value["Number"]) .. tostring(value["Operator"])] = true
                        end
                    end
                end
            end
        end))
        table.insert(self.combatCallbacks, AddCallback(self.player, "InCombatChanged", function()
            if not(self.player:IsInCombat()) then
                for key, value in pairs(data["Target_HealthThreshold"]) do
                    valuebar.targetHealthThresholdCooldowns[key .. tostring(value["Number"]) .. tostring(value["Operator"])] = Turbine.Engine.GetGameTime() - 1000
                    valuebar.targetHealthThresholdPercentage[key .. tostring(value["Number"]) .. tostring(value["Operator"])] = true
                end 
            end
        end))
    end

    if tablelength(data["Skills"]) > 0 then
        self.skills = true
        for key, value in pairs(data["Skills"]) do
            if SkillEventHandler:SkillExists(key) then
                -- Add the callback for this skill slot
                if not(self.skillCallbacks[key]) then self.skillCallbacks[key] = {} end
                table.insert(self.skillCallbacks[key], AddCallback(SkillEventHandler, key, function(skill) 
                    NoteHandler:HanldeEvent(data["Skills"][key])
                end))
            end
        end
    end

    table.insert(self.UI, valuebar)
end

function PrimeUI:BalladBar(data, parent)
    local x = self:ContentAlignement(data["Alignment"], (data["Width"] + 2 * data["BorderSize"]) * data["Scale"])
    local y = self.currentHeight
    if self.groupMode then
        x = data["X"]
        y = data["Y"]
    end

    local valuebar = BalladBarPrimeUIControl(parent, x, y, data["Width"], data["Height"], data["Scale"], data["Box Offset"], data["BorderSize"],
    self:ColourTupleToTurbineColour(data["DamageColour"]), self:ColourTupleToTurbineColour(data["HealColour"]), self:ColourTupleToTurbineColour(data["PowerColour"]), self:ColourTupleToTurbineColour(data["BackgroundColour"]), self:ColourTupleToTurbineColour(data["BorderColour"]), self:ColourTupleToTurbineColour(data["BoxBackgroundColour"]))

    if self.groupMode then
        valuebar.moveBox = MoveBox(valuebar, data["Width"] * data["Scale"], data["Height"] * data["Scale"], "Skill label")
        valuebar.savePosition = function()
            data["X"], data["Y"] = valuebar:GetPosition()
        end
    end

    
    valuebar.balladTypes = {
    	-- Red
    	["Ballad Damage Bonus"] = 1,

    	-- Green
    	["Ballad Healing Bonus"] = 2,

    	-- Blue
    	["Ballad Power Bonus"] = 3,
    }

    valuebar.effectFunction = function(effect)
        if valuebar.balladTypes[effect:GetName()] then
            valuebar:AddBallad(effect)
        end
    end

    for i = 1, self.player:GetEffects():GetCount() do
        valuebar.effectFunction(self.player:GetEffects():Get(i))
    end

    table.insert(self.effectAddedCallbacks, AddCallback(PlayerEffectsHandler, "EffectAdded", valuebar.effectFunction))

    table.insert(self.effectRemovedCallbacks, AddCallback(PlayerEffectsHandler, "EffectRemoved", 
    function(effect)
        if valuebar.balladTypes[effect:GetName()] then
            valuebar:RemoveBallad(effect:GetID())
        end
    end))

    self.currentHeight = self.currentHeight + data["Height"]

    table.insert(self.UI, valuebar)
end

function PrimeUI:SkillLabel(data, parent)
    local x = self:ContentAlignement(data["Alignment"], (data["Width"] + 2 * data["BorderSize"]) * data["Scale"])
    local y = self.currentHeight
    if self.groupMode then
        x = data["X"]
        y = data["Y"]
    end

    local valuebar = PrimeUISkillLabel(parent, x, y, data["Width"], data["Height"], data["Scale"], data["Icon"], self:ColourTupleToTurbineColour(data["BorderColour"]), data["BorderSize"], data["Clockwise"], data["Visibility_Threshold"], data["Hide_While_Not_Active"], data["Hide_While_Active"], data["Opacity_While_Active"], data["Opacity_While_Not_Active"], self:ColourTupleToTurbineColour(data["Highlight_Colour"]), data["Highlight_Size"], data["HighlightWhileOff"], data["Highlight"])

    if self.groupMode then
        valuebar.moveBox = MoveBox(valuebar, data["Width"] * data["Scale"], data["Height"] * data["Scale"], "Skill label")
        valuebar.savePosition = function()
            data["X"], data["Y"] = valuebar:GetPosition()
        end
    end

    valuebar.SkillID = {}
    valuebar.UniqueNames = {}

    if tablelength(data["Highlight_On_Effect"]) > 0 then
        table.insert(self.effectAddedCallbacks, AddCallback(PlayerEffectsHandler, "EffectAdded", 
        function(effect)
            local name = effect:GetName()
            if data["Highlight_On_Effect"][name] then
                local value = data["Highlight_On_Effect"][name]

                --????
                --if value.Fit_Icon then
                --    if data["Highlight_On_Effect"] ~= effect:GetIcon() then return end
                --end

                local ID = effect:GetID()
                valuebar.SkillID[ID] = true
                valuebar:ToggleHighlight(not(value["Invert_Highlight"]))
            end
        end))

        table.insert(self.effectRemovedCallbacks, AddCallback(PlayerEffectsHandler, "EffectRemoved", 
        function(effect)
            local ID = effect:GetID()
            if valuebar.SkillID[ID] then
                valuebar:ToggleHighlight(data["Highlight_On_Effect"][effect:GetName()]["Invert_Highlight"])
            end

        end))
    end

    if tablelength(data["EffectAdded"]) > 0 then
        table.insert(self.effectAddedCallbacks, AddCallback(PlayerEffectsHandler, "EffectAdded", 
        function(effect)
            local name = effect:GetName()
            if data["EffectAdded"][name] ~= nil then
                local time = data["EffectAdded"][name]
                local maxTime = data["EffectAdded"][name]
                if time == true then
                    time = effect:GetDuration() - (Turbine.Engine.GetGameTime() - effect:GetStartTime())
                    maxTime = effect:GetDuration()
                end
                valuebar:SetTimer(time, maxTime)
            end
        end))
    end

    if tablelength(data["EffectAdded_Target"]) > 0 then
        table.insert(self.effectTargetAddedCallbacks, TargetHandlerObject:AddEffectCallback( 
        function(effect)
            local name = effect:GetName()
            if data["EffectAdded_Target"][name] ~= nil then
                local time = data["EffectAdded_Target"][name]
                local maxTime = data["EffectAdded_Target"][name]
                if time == true then
                    time = effect:GetDuration() - (Turbine.Engine.GetGameTime() - effect:GetStartTime())
                    maxTime = effect:GetDuration()
                end
                valuebar:SetTimer(time, maxTime)
            end
        end))
    end

    if tablelength(data["EffectAdded_Group"]) > 0 then
        table.insert(self.effectAddedGroupCallbacks, AddCallback(GroupEffectsHandler, "EffectAdded", 
        function(args)
            local effect = args.effect
            local name = effect:GetName()
            if data["EffectAdded_Group"][name] ~= nil then
                local time = data["EffectAdded_Group"][name]
                local maxTime = data["EffectAdded_Group"][name]
                if time == true then
                    time = effect:GetDuration() - (Turbine.Engine.GetGameTime() - effect:GetStartTime())
                    maxTime = effect:GetDuration()
                end
                valuebar:SetTimer(time, maxTime)
            end
        end))
    end

    if tablelength(data["EffectRemoved"]) > 0 then
        table.insert(self.effectRemovedCallbacks, AddCallback(PlayerEffectsHandler, "EffectRemoved", 
        function(effect)
            local name = effect:GetName()
            if data["EffectRemoved"][name] ~= nil then
                local time = data["EffectRemoved"][name]
                local maxTime = data["EffectRemoved"][name]
                if time == true then
                    time = effect:GetDuration() - (Turbine.Engine.GetGameTime() - effect:GetStartTime())
                    maxTime = effect:GetDuration()
                end
                valuebar:SetTimer(time, maxTime)
            end
        end))
    end

    if tablelength(data["EnemyHits"]) > 0 then
        table.insert(self.enemyChatCallback, AddCallback(ChatEventHandler, "EnemyCombat", function(args)
            local value = data["EnemyHits"][args.SkillName]
            if value then
                valuebar:SetTimer(value, value)
            end
        end))

    end

    if tablelength(data["PlayerHits"]) > 0 then
        table.insert(self.playerChatCallback, AddCallback(ChatEventHandler, "PlayerCombat", function(args)
            local value = data["PlayerHits"][args.SkillName]
            if value then
                valuebar:SetTimer(value, value)
            end
        end))
    end

    if tablelength(data["PlayerBenefit"]) > 0 then
        table.insert(self.playerBenefitCallbacks, AddCallback(ChatEventHandler, "PlayerBenefit", function(args)
            local value = data["PlayerBenefit"][args.SkillName]
            if value then
                valuebar:SetTimer(value, value)
            end
        end))
    end

    if tablelength(data["PlayerHeal"]) > 0 then
        table.insert(self.playerHealCallbacks, AddCallback(ChatEventHandler, "PlayerHeal", function(args)
            local value = data["PlayerHeal"][args.SkillName]
            if value then
                valuebar:SetTimer(value, value)
            end
        end))
    end

    if tablelength(data["StandardChat"]) > 0 then
        table.insert(self.standardChatCallback, AddCallback(ChatEventHandler, "Standard", function(message)
            if data["StandardChat"][message] then
                valuebar:SetTimer(data["StandardChat"][message], data["StandardChat"][message])
            end
        end))
    end

    if tablelength(data["PartyChat"]) > 0 then
        table.insert(self.partyChatCallback, AddCallback(ChatEventHandler, "Party", function(message)
            if data["PartyChat"][message] then
                valuebar:SetTimer(data["PartyChat"][message], data["PartyChat"][message])
            end
        end))
    end

    if tablelength(data["SayChat"]) > 0 then
        table.insert(self.sayChatCallback, AddCallback(ChatEventHandler, "SayLine", function(message)
            if data["SayChat"][message]  then
                valuebar:SetTimer(data["SayChat"][message], data["SayChat"][message])
            end
        end))
    end


    if tablelength(data["Skills"]) > 0 then
        self.skills = true
        for key, value in pairs(data["Skills"]) do
            if SkillEventHandler:SkillExists(key) then
                -- Add the callback for this skill slot
                if not(self.skillCallbacks[key]) then self.skillCallbacks[key] = {} end
                table.insert(self.skillCallbacks[key], AddCallback(SkillEventHandler, key, function(skill) 
                    local time = data["Skills"][key]
                    local maxTime = data["Skills"][key]
                    if data["Skills"][key] == true then
                        time = skill:GetResetTime() - Turbine.Engine.GetGameTime()
                        maxTime = skill:GetCooldown()
                    end

                    valuebar:SetTimer(time, maxTime)
                end))
            end
        end
    end

    table.insert(self.UI, valuebar)
end

function PrimeUI:AlignmentToNumber(text)
    if text == "Middle" then return 2 end
    if text == "Left" then return 1 end
    if text == "Right" then return 3 end
    return 4
end

function PrimeUI:AlignmentToNumberSmall(text)
    if text == "Middle" then return 1 end
    if text == "Left" then return 2 end
    if text == "Right" then return 3 end
    return 3
end

function PrimeUI:ColourTupleToTurbineColour(tuple)
    if tuple then
        if type(tuple[1]) == "table" then
            local colourTable = {}
            for key, value in pairs(tuple) do
                if value then
                    colourTable[key] = Turbine.UI.Color(value[1], value[2], value[3], value[4])
                else
                    colourTable[key] = nil
                end
            end

            return colourTable
        else
            return Turbine.UI.Color(tuple[1], tuple[2], tuple[3], tuple[4])
        end
    else
        return Turbine.UI.Color(0,0,0,0)
    end
end

function PrimeUI:FontToFont(font)
    return Turbine.UI.Lotro.Font[font]
end

function PrimeUI:UnloadObject()
    for name, callbackTable in pairs(self.skillCallbacks) do
        for key, callback in pairs(callbackTable) do
            RemoveCallback(SkillEventHandler, name, callback)
            self.skillCallbacks[name][key] = nil
        end
        self.skillCallbacks[name] = nil
    end

    for key, callbackPair in pairs(self.playerCallbacks) do
        RemoveCallback(self.player, callbackPair[1], callbackPair[2])
        self.playerCallbacks[key] = nil
    end

    for key, callback in pairs(self.effectAddedCallbacks) do
        RemoveCallback(PlayerEffectsHandler, "EffectAdded", callback)
        self.effectAddedCallbacks[key] = nil
    end
    
    for key, callback in pairs(self.effectRemovedCallbacks) do
        RemoveCallback(PlayerEffectsHandler, "EffectRemoved", callback)
        self.effectRemovedCallbacks[key] = nil
    end

    for key, callback in pairs(self.effectAddedGroupCallbacks) do
        RemoveCallback(GroupEffectsHandler, "EffectAdded", callback)
        self.effectAddedGroupCallbacks[key] = nil
    end

    for key, value in pairs(self.enemyChatCallback) do
        RemoveCallback(ChatEventHandler, "EnemyCombat", value)
        self.enemyChatCallback[key] = nil
    end

    for key, value in pairs(self.playerChatCallback) do
        RemoveCallback(ChatEventHandler, "PlayerCombat", value)
        self.playerChatCallback[key] = nil
    end

    for key, value in pairs(self.playerHealCallbacks) do
        RemoveCallback(ChatEventHandler, "PlayerHeal", value)
        self.playerHealCallbacks[key] = nil
    end

    for key, value in pairs(self.playerBenefitCallbacks) do
        RemoveCallback(ChatEventHandler, "PlayerBenefit", value)
        self.playerBenefitCallbacks[key] = nil
    end
    
    for key, value in pairs(self.sayChatCallback) do
        RemoveCallback(ChatEventHandler, "SayLine", value)
        self.sayChatCallback[key] = nil
    end
    
    for key, value in pairs(self.partyChatCallback) do
        RemoveCallback(ChatEventHandler, "Party", value)
        self.partyChatCallback[key] = nil
    end
    
    for key, value in pairs(self.standardChatCallback) do
        RemoveCallback(ChatEventHandler, "Standard", value)
        self.standardChatCallback[key] = nil
    end

    for key, value in pairs(self.deathChatCallbacks) do
        RemoveCallback(ChatEventHandler, "Death", value)
        self.deathChatCallbacks[key] = nil
    end

    for key, value in pairs(self.instanceEventCallbacks) do
        RemoveCallback(InstanceController, "InstanceEvent", value)
        self.instanceEventCallbacks[key] = nil
    end

    for key, value in pairs(self.fightEventStartCallbacks) do
        RemoveCallback(InstanceController, "FightEventStart", value)
        self.fightEventStartCallbacks[key] = nil
    end

    
    for key, value in pairs(self.fightEventEndCallbacks) do
        RemoveCallback(InstanceController, "FightEventEnd", value)
        self.fightEventEndCallbacks[key] = nil
    end

    for key, value in pairs(self.effectTargetAddedCallbacks) do
        TargetHandlerObject:RemoveEffectCallback(value)
        self.effectTargetAddedCallbacks[key] = nil
    end

    for key, value in pairs(self.healthTargetCallbacks) do
        TargetHandlerObject:RemoveHealthCallback(value)
        self.healthTargetCallbacks[key] = nil
    end

    for key, value in pairs(self.temporaryHealthTargetCallbacks) do
        TargetHandlerObject:RemoveTemporaryHealthCallback(value)
        self.temporaryHealthTargetCallbacks[key] = nil
    end

    for key, value in pairs(self.powerTargetCallbacks) do
        TargetHandlerObject:RemovePowerCallback(value)
        self.powerTargetCallbacks[key] = nil
    end

    for key, value in pairs(self.targetInfoCallbacks) do
        TargetHandlerObject:RemoveTargetInfoCallback(value)
        self.targetInfoCallbacks[key] = nil
    end

    
    for key, callback in pairs(self.combatCallbacks) do
        RemoveCallback(self.player, "InCombatChanged", callback)
        self.playerCallbacks[key] = nil
    end
    
    for key, tool in pairs(self.tools) do
        tool:Unload()
        tool:SetVisible(false)
        self.tools[key] = nil
    end

    for key, UI in pairs(self.UI) do
        if UI.SetStretchMode then UI:SetStretchMode(nil) end
        UI:SetParent(nil)
        UI:SetVisible(false)
        self.UI[key] = nil
    end

    if self.moveBox then 
        self.moveBox:Unload() 
        self.moveBox = nil
    end

    if self.updateCallback then RemoveCallback(Updater, "ExecuteUpdate", self.updateCallback) end
    self.updateCallback = nil

    self.currentHeight = 0

    self.skillCallbacks = nil
    self.playerCallbacks = nil
    self.attrCallbacks = nil
    self.effectAddedCallbacks = nil
    self.effectRemovedCallbacks = nil
    self.playerChatCallback = nil
    self.enemyChatCallback = nil
    self.sayChatCallback = nil
    self.partyChatCallback = nil
    self.standardChatCallback = nil
    self.timers = nil
    self.icons = nil
    self.effectAddedGroupCallbacks = nil
    self.deathChatCallbacks = nil
    self.instanceEventCallbacks = nil
    self.fightEventStartCallbacks = nil
    self.fightEventEndCallbacks = nil
    self.effectTargetAddedCallbacks = nil
    self.healthTargetCallbacks = nil
    self.temporaryHealthTargetCallbacks = nil
    self.powerTargetCallbacks = nil
    self.targetInfoCallbacks = nil
    self.combatCallbacks = nil
    self.playerBenefitCallbacks = nil
    self.playerHealCallbacks = nil

    local controlList = self:GetControls()
    for i = 1, controlList:GetCount() do
        controlList:RemoveAt(1)
    end

    collectgarbage()

    self.timers = {}
    self.icons = {}
    self.skillCallbacks = {}
    self.playerCallbacks = {}
    self.attrCallbacks = {}
    self.effectAddedCallbacks = {}
    self.effectRemovedCallbacks = {}
    self.playerChatCallback = {}
    self.enemyChatCallback = {}
    self.sayChatCallback = {}
    self.partyChatCallback = {}
    self.standardChatCallback = {}
    self.effectAddedGroupCallbacks = {}
    self.deathChatCallbacks = {}
    self.instanceEventCallbacks = {}
    self.fightEventStartCallbacks = {}
    self.fightEventEndCallbacks = {}
    self.effectTargetAddedCallbacks = {}
    self.healthTargetCallbacks = {}
    self.temporaryHealthTargetCallbacks = {}
    self.powerTargetCallbacks = {}
    self.targetInfoCallbacks = {}
    self.combatCallbacks = {}
    self.playerBenefitCallbacks = {}
    self.playerHealCallbacks = {}
end